##############################################################################
#
# Copyright (c) 2004 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Unit tests for import / export contexts.
"""

import logging
import os
import tempfile
import time
import unittest
from io import BytesIO
from string import digits
from string import printable
from tarfile import TarFile
from tarfile import TarInfo

from DateTime.DateTime import DateTime
from OFS.Folder import Folder
from OFS.Image import File
from Testing.ZopeTestCase import ZopeTestCase

from .common import FilesystemTestBase
from .common import TarballTester
from .conformance import ConformsToIChunkableExportContext
from .conformance import ConformsToIChunkableImportContext
from .conformance import ConformsToIExportContext
from .conformance import ConformsToIImportContext
from .conformance import ConformsToISetupContext


printable_bytes = printable.encode('utf-8')
digits_bytes = digits.encode('utf-8')


class DummySite(Folder):

    pass


class DummyTool(Folder):

    pass


class DummyPdataStreamIterator:

    pass


class DirectoryImportContextTests(FilesystemTestBase, ConformsToISetupContext,
                                  ConformsToIImportContext,
                                  ConformsToIChunkableImportContext):

    _PROFILE_PATH = tempfile.mkdtemp(prefix='ICTTexts')

    def _getTargetClass(self):

        from ..context import DirectoryImportContext
        return DirectoryImportContext

    def test_getLogger(self):

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)
        self.assertEqual(len(ctx.listNotes()), 0)

        logger = ctx.getLogger('foo')
        logger.info('bar')

        self.assertEqual(len(ctx.listNotes()), 1)
        level, component, message = ctx.listNotes()[0]
        self.assertEqual(level, logging.INFO)
        self.assertEqual(component, 'foo')
        self.assertEqual(message, 'bar')

        ctx.clearNotes()
        self.assertEqual(len(ctx.listNotes()), 0)

    def test_readDataFile_nonesuch(self):

        FILENAME = 'nonesuch.txt'

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        self.assertEqual(ctx.readDataFile(FILENAME), None)

    def test_readDataFile_simple(self):

        FILENAME = 'simple.txt'
        self._makeFile(FILENAME, printable_bytes)

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        self.assertEqual(ctx.readDataFile(FILENAME), printable_bytes)

    def test_readDataFile_subdir(self):

        FILENAME = 'subdir/nested.txt'
        self._makeFile(FILENAME, printable_bytes)

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        self.assertEqual(ctx.readDataFile(FILENAME), printable_bytes)

    def test_getLastModified_nonesuch(self):
        FILENAME = 'nonesuch.txt'
        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        self.assertEqual(ctx.getLastModified(FILENAME), None)

    def test_getLastModified_simple(self):

        FILENAME = 'simple.txt'
        fqpath = self._makeFile(FILENAME, printable_bytes)
        WHEN = os.path.getmtime(fqpath)
        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        lm = ctx.getLastModified(FILENAME)
        self.assertIsInstance(lm, DateTime)
        self.assertEqual(lm, DateTime(WHEN))

    def test_getLastModified_subdir(self):

        FILENAME = 'subdir.txt'
        SUBDIR = 'subdir'
        PATH = os.path.join(SUBDIR, FILENAME)
        fqpath = self._makeFile(PATH, printable_bytes)
        WHEN = os.path.getmtime(fqpath)
        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        lm = ctx.getLastModified(PATH)
        self.assertIsInstance(lm, DateTime)
        self.assertEqual(lm, DateTime(WHEN))

    def test_getLastModified_directory(self):

        FILENAME = 'subdir.txt'
        SUBDIR = 'subdir'
        PATH = os.path.join(SUBDIR, FILENAME)
        fqpath = self._makeFile(PATH, printable_bytes)
        path, file = os.path.split(fqpath)
        WHEN = os.path.getmtime(path)
        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        lm = ctx.getLastModified(SUBDIR)
        self.assertIsInstance(lm, DateTime)
        self.assertEqual(lm, DateTime(WHEN))

    def test_isDirectory_nonesuch(self):

        FILENAME = 'nonesuch.txt'

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        self.assertEqual(ctx.isDirectory(FILENAME), None)

    def test_isDirectory_simple(self):

        FILENAME = 'simple.txt'
        self._makeFile(FILENAME, printable_bytes)

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        self.assertEqual(ctx.isDirectory(FILENAME), False)

    def test_isDirectory_nested(self):

        SUBDIR = 'subdir'
        FILENAME = os.path.join(SUBDIR, 'nested.txt')
        self._makeFile(FILENAME, printable_bytes)

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        self.assertEqual(ctx.isDirectory(FILENAME), False)

    def test_isDirectory_directory(self):

        SUBDIR = 'subdir'
        FILENAME = os.path.join(SUBDIR, 'nested.txt')
        self._makeFile(FILENAME, printable_bytes)

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        self.assertEqual(ctx.isDirectory(SUBDIR), True)

    def test_listDirectory_nonesuch(self):

        FILENAME = 'nonesuch.txt'

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        self.assertEqual(ctx.listDirectory(FILENAME), None)

    def test_listDirectory_root(self):

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        FILENAME = 'simple.txt'
        self._makeFile(FILENAME, printable_bytes)

        self.assertEqual(len(ctx.listDirectory(None)), 1)
        self.assertIn(FILENAME, ctx.listDirectory(None))

    def test_listDirectory_simple(self):

        FILENAME = 'simple.txt'
        self._makeFile(FILENAME, printable_bytes)

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        self.assertEqual(ctx.listDirectory(FILENAME), None)

    def test_listDirectory_nested(self):

        SUBDIR = 'subdir'
        FILENAME = os.path.join(SUBDIR, 'nested.txt')
        self._makeFile(FILENAME, printable_bytes)

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        self.assertEqual(ctx.listDirectory(FILENAME), None)

    def test_listDirectory_single(self):

        SUBDIR = 'subdir'
        FILENAME = os.path.join(SUBDIR, 'nested.txt')
        self._makeFile(FILENAME, printable_bytes)

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        names = ctx.listDirectory(SUBDIR)
        self.assertEqual(len(names), 1)
        self.assertIn('nested.txt', names)

    def test_listDirectory_multiple(self):

        SUBDIR = 'subdir'
        FILENAME = os.path.join(SUBDIR, 'nested.txt')
        self._makeFile(FILENAME, printable_bytes)
        self._makeFile(os.path.join(SUBDIR, 'another.txt'), b'ABC')

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        names = ctx.listDirectory(SUBDIR)
        self.assertEqual(len(names), 2)
        self.assertIn('nested.txt', names)
        self.assertIn('another.txt', names)

    def test_listDirectory_skip_implicit(self):

        SUBDIR = 'subdir'
        FILENAME = os.path.join(SUBDIR, 'nested.txt')
        self._makeFile(FILENAME, printable_bytes)
        self._makeFile(os.path.join(SUBDIR, 'another.txt'), b'ABC')
        self._makeFile(os.path.join(SUBDIR, 'another.txt~'), b'123')
        self._makeFile(os.path.join(SUBDIR, 'CVS/skip.txt'), b'DEF')
        self._makeFile(os.path.join(SUBDIR, '.svn/skip.txt'), b'GHI')

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        names = ctx.listDirectory(SUBDIR)
        self.assertEqual(len(names), 2)
        self.assertIn('nested.txt', names)
        self.assertIn('another.txt', names)
        self.assertNotIn('another.txt~', names)
        self.assertNotIn('CVS', names)
        self.assertNotIn('.svn', names)

    def test_listDirectory_skip_explicit(self):

        SUBDIR = 'subdir'
        FILENAME = os.path.join(SUBDIR, 'nested.txt')
        self._makeFile(FILENAME, printable_bytes)
        self._makeFile(os.path.join(SUBDIR, 'another.txt'), b'ABC')
        self._makeFile(os.path.join(SUBDIR, 'another.bak'), b'123')
        self._makeFile(os.path.join(SUBDIR, 'CVS/skip.txt'), b'DEF')
        self._makeFile(os.path.join(SUBDIR, '.svn/skip.txt'), b'GHI')

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        names = ctx.listDirectory(SUBDIR, skip=('nested.txt',),
                                  skip_suffixes=('.bak',))
        self.assertEqual(len(names), 3)
        self.assertNotIn('nested.txt', names)
        self.assertNotIn('nested.bak', names)
        self.assertIn('another.txt', names)
        self.assertIn('CVS', names)
        self.assertIn('.svn', names)


class DirectoryExportContextTests(FilesystemTestBase, ConformsToISetupContext,
                                  ConformsToIExportContext,
                                  ConformsToIChunkableExportContext):

    _PROFILE_PATH = tempfile.mkdtemp(prefix='ECTTexts')

    def _getTargetClass(self):

        from ..context import DirectoryExportContext
        return DirectoryExportContext

    def test_getLogger(self):

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)
        self.assertEqual(len(ctx.listNotes()), 0)

        logger = ctx.getLogger('foo')
        logger.info('bar')

        self.assertEqual(len(ctx.listNotes()), 1)
        level, component, message = ctx.listNotes()[0]
        self.assertEqual(level, logging.INFO)
        self.assertEqual(component, 'foo')
        self.assertEqual(message, 'bar')

        ctx.clearNotes()
        self.assertEqual(len(ctx.listNotes()), 0)

    def test_writeDataFile_simple(self):

        FILENAME = 'simple.txt'
        fqname = self._makeFile(FILENAME, printable_bytes)

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        ctx.writeDataFile(FILENAME, digits_bytes, 'text/plain')

        with open(fqname, 'rb') as fp:
            self.assertEqual(fp.read(), digits_bytes)

    def test_writeDataFile_unicode(self):

        text = 'Kein Weltraum links vom Ger\xe4t'
        FILENAME = 'unicode.txt'
        fqname = self._makeFile(FILENAME, printable_bytes)

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        ctx.writeDataFile(FILENAME, text, 'text/plain')

        with open(fqname, 'rb') as fp:
            self.assertEqual(fp.read(), text.encode('UTF-8'))

    def test_writeDataFile_new_subdir(self):

        SUBDIR = 'subdir'
        FILENAME = 'nested.txt'
        fqname = os.path.join(self._PROFILE_PATH, SUBDIR, FILENAME)

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        ctx.writeDataFile(FILENAME, digits_bytes, 'text/plain', SUBDIR)

        with open(fqname, 'rb') as fp:
            self.assertEqual(fp.read(), digits_bytes)

    def test_writeDataFile_overwrite(self):

        SUBDIR = 'subdir'
        FILENAME = 'nested.txt'
        fqname = self._makeFile(os.path.join(SUBDIR, FILENAME),
                                printable_bytes)

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        ctx.writeDataFile(FILENAME, digits_bytes, 'text/plain', SUBDIR)

        with open(fqname, 'rb') as fp:
            self.assertEqual(fp.read(), digits_bytes)

    def test_writeDataFile_existing_subdir(self):

        SUBDIR = 'subdir'
        FILENAME = 'nested.txt'
        self._makeFile(os.path.join(SUBDIR, 'another.txt'), printable_bytes)
        fqname = os.path.join(self._PROFILE_PATH, SUBDIR, FILENAME)

        site = DummySite('site').__of__(self.app)
        ctx = self._makeOne(site, self._PROFILE_PATH)

        ctx.writeDataFile(FILENAME, digits_bytes, 'text/plain', SUBDIR)

        with open(fqname, 'rb') as fp:
            self.assertEqual(fp.read(), digits_bytes)


class TarballImportContextTests(ZopeTestCase, ConformsToISetupContext,
                                ConformsToIImportContext):

    def _getTargetClass(self):

        from ..context import TarballImportContext
        return TarballImportContext

    def _makeOne(self, file_dict={}, mod_time=None, *args, **kw):

        archive_stream = BytesIO()
        archive = TarFile.open('test.tar.gz', 'w:gz', archive_stream)

        def _addOneMember(path, data, modtime):
            stream = BytesIO(v)
            info = TarInfo(k)
            info.size = len(v)
            info.mtime = modtime
            archive.addfile(info, stream)

        def _addMember(filename, data, modtime):
            from tarfile import DIRTYPE
            parents = filename.split('/')[:-1]
            while parents:
                path = '/'.join(parents) + '/'
                if path not in archive.getnames():
                    info = TarInfo(path)
                    info.type = DIRTYPE
                    info.mtime = modtime
                    archive.addfile(info)
                parents.pop()
            _addOneMember(filename, data, modtime)

        file_items = file_dict.items() or [(
            'dummy', b'')]  # empty archive barfs

        if mod_time is None:
            mod_time = time.time()

        for k, v in file_items:
            if isinstance(v, str):
                v = v.encode('utf-8')
            _addMember(k, v, mod_time)

        archive.close()
        bits = archive_stream.getvalue()

        site = DummySite('site').__of__(self.app)
        site._setObject('setup_tool', Folder('setup_tool'))
        tool = site._getOb('setup_tool')

        ctx = self._getTargetClass()(tool, bits, *args, **kw)

        return site, tool, ctx.__of__(tool)

    def test_getLogger(self):

        site, tool, ctx = self._makeOne()
        self.assertEqual(len(ctx.listNotes()), 0)

        logger = ctx.getLogger('foo')
        logger.info('bar')

        self.assertEqual(len(ctx.listNotes()), 1)
        level, component, message = ctx.listNotes()[0]
        self.assertEqual(level, logging.INFO)
        self.assertEqual(component, 'foo')
        self.assertEqual(message, 'bar')

        ctx.clearNotes()
        self.assertEqual(len(ctx.listNotes()), 0)

    def test_ctorparms(self):

        ENCODING = 'latin-1'
        site, tool, ctx = self._makeOne(encoding=ENCODING, should_purge=True)

        self.assertEqual(ctx.getEncoding(), ENCODING)
        self.assertEqual(ctx.shouldPurge(), True)

    def test_empty(self):

        site, tool, ctx = self._makeOne()

        self.assertEqual(ctx.getSite(), site)
        self.assertEqual(ctx.getEncoding(), None)
        self.assertEqual(ctx.shouldPurge(), False)

        # These methods are all specified to return 'None' for non-existing
        # paths / entities
        self.assertEqual(ctx.isDirectory('nonesuch/path'), None)
        self.assertEqual(ctx.listDirectory('nonesuch/path'), None)

    def test_readDataFile_nonesuch(self):

        FILENAME = 'nonesuch.txt'

        site, tool, ctx = self._makeOne()

        self.assertEqual(ctx.readDataFile(FILENAME), None)
        self.assertEqual(ctx.readDataFile(FILENAME, 'subdir'), None)

    def test_readDataFile_simple(self):

        FILENAME = 'simple.txt'

        site, tool, ctx = self._makeOne({FILENAME: printable_bytes})

        self.assertEqual(ctx.readDataFile(FILENAME), printable_bytes)

    def test_readDataFile_subdir(self):

        FILENAME = 'subdir.txt'
        SUBDIR = 'subdir'

        site, tool, ctx = self._makeOne({f'{SUBDIR}/{FILENAME}':
                                         printable_bytes})

        self.assertEqual(ctx.readDataFile(FILENAME, SUBDIR), printable_bytes)

    def test_getLastModified_nonesuch(self):
        FILENAME = 'nonesuch.txt'
        ctx = self._makeOne()[2]

        self.assertEqual(ctx.getLastModified(FILENAME), None)

    def test_getLastModified_simple(self):

        FILENAME = 'simple.txt'
        WHEN = 999999999.0
        ctx = self._makeOne({FILENAME: printable_bytes}, mod_time=WHEN)[2]

        lm = ctx.getLastModified(FILENAME)
        self.assertIsInstance(lm, DateTime)
        self.assertEqual(lm, DateTime(WHEN))

    def test_getLastModified_subdir(self):

        FILENAME = 'subdir.txt'
        SUBDIR = 'subdir'
        PATH = f'{SUBDIR}/{FILENAME}'
        WHEN = 999999999.0
        ctx = self._makeOne({PATH: printable_bytes}, mod_time=WHEN)[2]

        lm = ctx.getLastModified(PATH)
        self.assertIsInstance(lm, DateTime)
        self.assertEqual(lm, DateTime(WHEN))

    def test_getLastModified_directory(self):

        FILENAME = 'subdir.txt'
        SUBDIR = 'subdir'
        PATH = f'{SUBDIR}/{FILENAME}'
        WHEN = 999999999.0
        ctx = self._makeOne({PATH: printable_bytes}, mod_time=WHEN)[2]

        lm = ctx.getLastModified(SUBDIR)
        self.assertIsInstance(lm, DateTime)
        self.assertEqual(lm, DateTime(WHEN))

    def test_isDirectory_nonesuch(self):

        FILENAME = 'nonesuch.txt'

        site, tool, ctx = self._makeOne()

        self.assertEqual(ctx.isDirectory(FILENAME), None)

    def test_isDirectory_simple(self):

        FILENAME = 'simple.txt'

        site, tool, ctx = self._makeOne({FILENAME: printable_bytes})

        self.assertEqual(ctx.isDirectory(FILENAME), False)

    def test_isDirectory_nested(self):

        SUBDIR = 'subdir'
        FILENAME = 'nested.txt'
        PATH = f'{SUBDIR}/{FILENAME}'

        site, tool, ctx = self._makeOne({PATH: printable_bytes})

        self.assertEqual(ctx.isDirectory(PATH), False)

    def test_isDirectory_subdir(self):

        SUBDIR = 'subdir'
        FILENAME = 'nested.txt'
        PATH = f'{SUBDIR}/{FILENAME}'

        site, tool, ctx = self._makeOne({PATH: printable_bytes})

        self.assertEqual(ctx.isDirectory(SUBDIR), True)

    def test_listDirectory_nonesuch(self):

        SUBDIR = 'nonesuch/path'

        site, tool, ctx = self._makeOne()

        self.assertEqual(ctx.listDirectory(SUBDIR), None)

    def test_listDirectory_root(self):

        FILENAME = 'simple.txt'

        site, tool, ctx = self._makeOne({FILENAME: printable_bytes})

        self.assertEqual(len(ctx.listDirectory(None)), 1)
        self.assertIn(FILENAME, ctx.listDirectory(None))

    def test_listDirectory_simple(self):

        FILENAME = 'simple.txt'

        site, tool, ctx = self._makeOne({FILENAME: printable_bytes})

        self.assertEqual(ctx.listDirectory(FILENAME), None)

    def test_listDirectory_nested(self):

        SUBDIR = 'subdir'
        FILENAME = 'nested.txt'
        PATH = f'{SUBDIR}/{FILENAME}'

        site, tool, ctx = self._makeOne({PATH: printable_bytes})

        self.assertEqual(ctx.listDirectory(PATH), None)

    def test_listDirectory_single(self):

        SUBDIR = 'subdir'
        FILENAME = 'nested.txt'
        PATH = f'{SUBDIR}/{FILENAME}'

        site, tool, ctx = self._makeOne({PATH: printable_bytes})

        names = ctx.listDirectory(SUBDIR)
        self.assertEqual(len(names), 1)
        self.assertIn(FILENAME, names)

    def test_listDirectory_multiple(self):

        from string import ascii_uppercase

        SUBDIR = 'subdir'
        FILENAME1 = 'nested.txt'
        PATH1 = f'{SUBDIR}/{FILENAME1}'
        FILENAME2 = 'another.txt'
        PATH2 = f'{SUBDIR}/{FILENAME2}'

        site, tool, ctx = self._makeOne({PATH1: printable_bytes,
                                         PATH2: ascii_uppercase})

        names = ctx.listDirectory(SUBDIR)
        self.assertEqual(len(names), 2)
        self.assertIn(FILENAME1, names)
        self.assertIn(FILENAME2, names)

    def test_listDirectory_skip(self):

        from string import ascii_uppercase

        SUBDIR = 'subdir'
        FILENAME1 = 'nested.txt'
        PATH1 = f'{SUBDIR}/{FILENAME1}'
        FILENAME2 = 'another.txt'
        PATH2 = f'{SUBDIR}/{FILENAME2}'
        FILENAME3 = 'another.bak'
        PATH3 = f'{SUBDIR}/{FILENAME3}'

        site, tool, ctx = self._makeOne({PATH1: printable_bytes,
                                         PATH2: ascii_uppercase,
                                         PATH3: 'xyz'})

        names = ctx.listDirectory(SUBDIR, skip=(FILENAME1,),
                                  skip_suffixes=('.bak',))
        self.assertEqual(len(names), 1)
        self.assertNotIn(FILENAME1, names)
        self.assertIn(FILENAME2, names)
        self.assertNotIn(FILENAME3, names)


class TarballExportContextTests(ZopeTestCase, ConformsToISetupContext,
                                ConformsToIExportContext, TarballTester):

    def _getTargetClass(self):

        from ..context import TarballExportContext
        return TarballExportContext

    def test_getLogger(self):

        site = DummySite('site').__of__(self.app)
        ctx = self._getTargetClass()(site)

        self.assertEqual(len(ctx.listNotes()), 0)

        logger = ctx.getLogger('foo')
        logger.info('bar')

        self.assertEqual(len(ctx.listNotes()), 1)
        level, component, message = ctx.listNotes()[0]
        self.assertEqual(level, logging.INFO)
        self.assertEqual(component, 'foo')
        self.assertEqual(message, 'bar')

        ctx.clearNotes()
        self.assertEqual(len(ctx.listNotes()), 0)

    def test_writeDataFile_simple(self):

        now = int(time.time())

        site = DummySite('site').__of__(self.app)
        ctx = self._getTargetClass()(site)

        ctx.writeDataFile('foo.txt', printable_bytes, 'text/plain')

        fileish = BytesIO(ctx.getArchive())

        self._verifyTarballContents(fileish, ['foo.txt'], now)
        self._verifyTarballEntry(fileish, 'foo.txt', printable_bytes)

    def test_writeDataFile_umlauts(self):

        text = 'Kein Weltraum links vom Ger\xe4t'
        now = int(time.time())

        site = DummySite('site').__of__(self.app)
        ctx = self._getTargetClass()(site)

        ctx.writeDataFile('foo.txt', text, 'text/plain')

        fileish = BytesIO(ctx.getArchive())

        self._verifyTarballContents(fileish, ['foo.txt'], now)
        self._verifyTarballEntry(fileish, 'foo.txt', text.encode('UTF-8'))

    def test_writeDataFile_multiple(self):

        site = DummySite('site').__of__(self.app)
        ctx = self._getTargetClass()(site)

        ctx.writeDataFile('foo.txt', printable_bytes, 'text/plain')
        ctx.writeDataFile('bar.txt', digits_bytes, 'text/plain')

        fileish = BytesIO(ctx.getArchive())

        self._verifyTarballContents(fileish, ['foo.txt', 'bar.txt'])
        self._verifyTarballEntry(fileish, 'foo.txt', printable_bytes)
        self._verifyTarballEntry(fileish, 'bar.txt', digits_bytes)

    def test_writeDataFile_PdataStreamIterator(self):

        site = DummySite('site').__of__(self.app)
        ctx = self._getTargetClass()(site)

        fp = tempfile.TemporaryFile()
        fp.write(printable_bytes)
        fp.seek(0)
        pData = DummyPdataStreamIterator()
        pData.file = fp
        pData.size = len(printable_bytes)
        ctx.writeDataFile('foo.txt', pData, 'text/plain')

        fileish = BytesIO(ctx.getArchive())

        self._verifyTarballContents(fileish, ['foo.txt'])
        self._verifyTarballEntry(fileish, 'foo.txt', printable_bytes)

        fp.close()  # Prevent unclosed file warning

    def test_writeDataFile_subdir(self):

        site = DummySite('site').__of__(self.app)
        ctx = self._getTargetClass()(site)

        ctx.writeDataFile('foo.txt', printable_bytes, 'text/plain')
        ctx.writeDataFile('bar/baz.txt', digits_bytes, 'text/plain')

        fileish = BytesIO(ctx.getArchive())

        self._verifyTarballContents(fileish,
                                    ['foo.txt', 'bar', 'bar/baz.txt'])
        self._verifyTarballEntry(fileish, 'foo.txt', printable_bytes)
        self._verifyTarballEntry(fileish, 'bar/baz.txt', digits_bytes)


class SnapshotExportContextTests(ZopeTestCase, ConformsToISetupContext,
                                 ConformsToIExportContext):

    def _getTargetClass(self):

        from ..context import SnapshotExportContext
        return SnapshotExportContext

    def _makeOne(self, *args, **kw):

        return self._getTargetClass()(*args, **kw)

    def test_getLogger(self):

        site = DummySite('site').__of__(self.app)
        site.setup_tool = DummyTool('setup_tool')
        tool = site.setup_tool
        ctx = self._makeOne(tool, 'simple')

        self.assertEqual(len(ctx.listNotes()), 0)

        logger = ctx.getLogger('foo')
        logger.info('bar')

        self.assertEqual(len(ctx.listNotes()), 1)
        level, component, message = ctx.listNotes()[0]
        self.assertEqual(level, logging.INFO)
        self.assertEqual(component, 'foo')
        self.assertEqual(message, 'bar')

        ctx.clearNotes()
        self.assertEqual(len(ctx.listNotes()), 0)

    def test_writeDataFile_simple_image(self):

        from OFS.Image import Image
        FILENAME = 'simple.txt'
        CONTENT_TYPE = 'image/png'
        png_filename = os.path.join(os.path.split(__file__)[0], 'simple.png')
        png_file = open(png_filename, 'rb')
        png_data = png_file.read()
        png_file.close()

        site = DummySite('site').__of__(self.app)
        site.setup_tool = DummyTool('setup_tool')
        tool = site.setup_tool
        ctx = self._makeOne(tool, 'simple')

        ctx.writeDataFile(FILENAME, png_data, CONTENT_TYPE)

        snapshot = tool.snapshots._getOb('simple')

        self.assertEqual(len(snapshot.objectIds()), 1)
        self.assertIn(FILENAME, snapshot.objectIds())

        fileobj = snapshot._getOb(FILENAME)

        self.assertEqual(fileobj.getId(), FILENAME)
        self.assertEqual(fileobj.meta_type, Image.meta_type)
        self.assertEqual(fileobj.getContentType(), CONTENT_TYPE)
        self.assertEqual(fileobj.data, png_data)

    def test_writeDataFile_simple_plain_text(self):

        FILENAME = 'simple.txt'
        CONTENT_TYPE = 'text/plain'

        site = DummySite('site').__of__(self.app)
        site.setup_tool = DummyTool('setup_tool')
        tool = site.setup_tool
        ctx = self._makeOne(tool, 'simple')

        ctx.writeDataFile(FILENAME, digits_bytes, CONTENT_TYPE)

        snapshot = tool.snapshots._getOb('simple')

        self.assertEqual(len(snapshot.objectIds()), 1)
        self.assertIn(FILENAME, snapshot.objectIds())

        fileobj = snapshot._getOb(FILENAME)

        self.assertEqual(fileobj.getId(), FILENAME)
        self.assertEqual(fileobj.meta_type, File.meta_type)
        self.assertEqual(fileobj.getContentType(), CONTENT_TYPE)
        self.assertEqual(fileobj.data, digits_bytes)

    def test_writeDataFile_simple_plain_text_unicode(self):
        FILENAME = 'simple.txt'
        CONTENT_TYPE = 'text/plain'
        CONTENT = 'Unicode, with non-ASCII: %s.' % chr(150)

        site = DummySite('site').__of__(self.app)
        site.setup_tool = DummyTool('setup_tool')
        tool = site.setup_tool
        ctx = self._makeOne(tool, 'simple', 'latin_1')
        ctx.writeDataFile(FILENAME, CONTENT, CONTENT_TYPE)

        snapshot = tool.snapshots._getOb('simple')

        self.assertEqual(len(snapshot.objectIds()), 1)
        self.assertIn(FILENAME, snapshot.objectIds())

        fileobj = snapshot._getOb(FILENAME)

        self.assertEqual(fileobj.getId(), FILENAME)
        self.assertEqual(fileobj.meta_type, File.meta_type)
        self.assertEqual(fileobj.getContentType(), CONTENT_TYPE)
        self.assertEqual(fileobj.data, CONTENT.encode(ctx._encoding))

    def test_writeDataFile_simple_xml(self):

        from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate
        FILENAME = 'simple.xml'
        CONTENT_TYPE = 'text/xml'
        _XML = """<?xml version="1.0"?><simple />"""

        site = DummySite('site').__of__(self.app)
        site.setup_tool = DummyTool('setup_tool')
        tool = site.setup_tool
        ctx = self._makeOne(tool, 'simple')

        ctx.writeDataFile(FILENAME, _XML, CONTENT_TYPE)

        snapshot = tool.snapshots._getOb('simple')

        self.assertEqual(len(snapshot.objectIds()), 1)
        self.assertIn(FILENAME, snapshot.objectIds())

        template = snapshot._getOb(FILENAME)

        self.assertEqual(template.getId(), FILENAME)
        self.assertEqual(template.meta_type, ZopePageTemplate.meta_type)
        self.assertEqual(template.read(), _XML)
        self.assertFalse(template.html())

    def test_writeDataFile_subdir_dtml(self):

        from OFS.DTMLDocument import DTMLDocument
        FILENAME = 'simple.dtml'
        CONTENT_TYPE = 'text/html'
        _HTML = """<html><body><h1>HTML</h1></body></html>"""

        site = DummySite('site').__of__(self.app)
        site.setup_tool = DummyTool('setup_tool')
        tool = site.setup_tool
        ctx = self._makeOne(tool, 'simple')

        ctx.writeDataFile(FILENAME, _HTML, CONTENT_TYPE, 'sub1')

        snapshot = tool.snapshots._getOb('simple')
        sub1 = snapshot._getOb('sub1')

        self.assertEqual(len(sub1.objectIds()), 1)
        self.assertIn(FILENAME, sub1.objectIds())

        template = sub1._getOb(FILENAME)

        self.assertEqual(template.getId(), FILENAME)
        self.assertEqual(template.meta_type, DTMLDocument.meta_type)
        self.assertEqual(template.read(), _HTML)

        ctx.writeDataFile('sub1/%s2' % FILENAME, _HTML, CONTENT_TYPE)
        self.assertEqual(len(sub1.objectIds()), 2)

    def test_writeDataFile_nested_subdirs_html(self):

        from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate
        FILENAME = 'simple.html'
        CONTENT_TYPE = 'text/html'
        _HTML = """<html><body><h1>HTML</h1></body></html>"""

        site = DummySite('site').__of__(self.app)
        site.setup_tool = DummyTool('setup_tool')
        tool = site.setup_tool
        ctx = self._makeOne(tool, 'simple')

        ctx.writeDataFile(FILENAME, _HTML, CONTENT_TYPE, 'sub1/sub2')

        snapshot = tool.snapshots._getOb('simple')
        sub1 = snapshot._getOb('sub1')
        sub2 = sub1._getOb('sub2')

        self.assertEqual(len(sub2.objectIds()), 1)
        self.assertIn(FILENAME, sub2.objectIds())

        template = sub2._getOb(FILENAME)

        self.assertEqual(template.getId(), FILENAME)
        self.assertEqual(template.meta_type, ZopePageTemplate.meta_type)
        self.assertEqual(template.read(), _HTML)
        self.assertTrue(template.html())

    def test_writeDataFile_multiple(self):

        site = DummySite('site').__of__(self.app)
        site.setup_tool = DummyTool('setup_tool')
        tool = site.setup_tool
        ctx = self._makeOne(tool, 'multiple')

        ctx.writeDataFile('foo.txt', printable_bytes, 'text/plain')
        ctx.writeDataFile('bar.txt', digits_bytes, 'text/plain')

        snapshot = tool.snapshots._getOb('multiple')

        self.assertEqual(len(snapshot.objectIds()), 2)

        for id in ['foo.txt', 'bar.txt']:
            self.assertIn(id, snapshot.objectIds())


class SnapshotImportContextTests(ZopeTestCase, ConformsToISetupContext,
                                 ConformsToIImportContext):

    def _getTargetClass(self):

        from ..context import SnapshotImportContext
        return SnapshotImportContext

    def _makeOne(self, context_id, *args, **kw):

        site = DummySite('site').__of__(self.app)
        site._setObject('setup_tool', Folder('setup_tool'))
        tool = site._getOb('setup_tool')

        tool._setObject('snapshots', Folder('snapshots'))
        tool.snapshots._setObject(context_id, Folder(context_id))

        ctx = self._getTargetClass()(tool, context_id, *args, **kw)

        return site, tool, ctx.__of__(tool)

    def _makeFile(self, tool, snapshot_id, filename, contents,
                  content_type='text/plain', mod_time=None, subdir=None):

        snapshots = tool._getOb('snapshots')
        folder = snapshots._getOb(snapshot_id)

        if subdir is not None:

            for element in subdir.split('/'):

                try:
                    folder = folder._getOb(element)
                except AttributeError:
                    folder._setObject(element, Folder(element))
                    folder = folder._getOb(element)

        file = File(filename, '', contents, content_type)
        folder._setObject(filename, file)

        if mod_time is not None:
            mod_time = DateTime(mod_time).timeTime()
            folder._faux_mod_time = file._faux_mod_time = mod_time

        return folder._getOb(filename)

    def test_getLogger(self):

        SNAPSHOT_ID = 'note'
        site, tool, ctx = self._makeOne(SNAPSHOT_ID)

        self.assertEqual(len(ctx.listNotes()), 0)

        logger = ctx.getLogger('foo')
        logger.info('bar')

        self.assertEqual(len(ctx.listNotes()), 1)
        level, component, message = ctx.listNotes()[0]
        self.assertEqual(level, logging.INFO)
        self.assertEqual(component, 'foo')
        self.assertEqual(message, 'bar')

        ctx.clearNotes()
        self.assertEqual(len(ctx.listNotes()), 0)

    def test_ctorparms(self):

        SNAPSHOT_ID = 'ctorparms'
        ENCODING = 'latin-1'
        site, tool, ctx = self._makeOne(SNAPSHOT_ID, encoding=ENCODING,
                                        should_purge=True)

        self.assertEqual(ctx.getEncoding(), ENCODING)
        self.assertEqual(ctx.shouldPurge(), True)

    def test_empty(self):

        SNAPSHOT_ID = 'empty'
        site, tool, ctx = self._makeOne(SNAPSHOT_ID)

        self.assertEqual(ctx.getSite(), site)
        self.assertEqual(ctx.getEncoding(), None)
        self.assertEqual(ctx.shouldPurge(), False)

        # These methods are all specified to return 'None' for non-existing
        # paths / entities
        self.assertEqual(ctx.isDirectory('nonesuch/path'), None)
        self.assertEqual(ctx.listDirectory('nonesuch/path'), None)

    def test_readDataFile_nonesuch(self):

        SNAPSHOT_ID = 'readDataFile_nonesuch'
        FILENAME = 'nonesuch.txt'

        site, tool, ctx = self._makeOne(SNAPSHOT_ID)

        self.assertEqual(ctx.readDataFile(FILENAME), None)
        self.assertEqual(ctx.readDataFile(FILENAME, 'subdir'), None)

    def test_readDataFile_simple(self):

        SNAPSHOT_ID = 'readDataFile_simple'
        FILENAME = 'simple.txt'

        site, tool, ctx = self._makeOne(SNAPSHOT_ID)
        self._makeFile(tool, SNAPSHOT_ID, FILENAME, printable_bytes)

        self.assertEqual(ctx.readDataFile(FILENAME), printable_bytes)

    def test_readDataFile_Pdata(self):

        from OFS.Image import Pdata

        SNAPSHOT_ID = 'readDataFile_Pdata'
        FILENAME = 'pdata.txt'

        site, tool, ctx = self._makeOne(SNAPSHOT_ID)
        self._makeFile(tool, SNAPSHOT_ID, FILENAME, Pdata(printable_bytes))

        self.assertEqual(ctx.readDataFile(FILENAME), printable_bytes)

    def test_readDataFile_subdir(self):

        SNAPSHOT_ID = 'readDataFile_subdir'
        FILENAME = 'subdir.txt'
        SUBDIR = 'subdir'

        site, tool, ctx = self._makeOne(SNAPSHOT_ID)
        self._makeFile(tool, SNAPSHOT_ID, FILENAME, printable_bytes,
                       subdir=SUBDIR)

        self.assertEqual(ctx.readDataFile(FILENAME, SUBDIR), printable_bytes)
        self.assertEqual(ctx.readDataFile(f'{SUBDIR}/{FILENAME}'),
                         printable_bytes)

    def test_getLastModified_nonesuch(self):
        FILENAME = 'nonesuch.txt'
        SNAPSHOT_ID = 'getLastModified_nonesuch'
        ctx = self._makeOne(SNAPSHOT_ID)[2]

        self.assertEqual(ctx.getLastModified(FILENAME), None)

    def test_getLastModified_simple(self):

        FILENAME = 'simple.txt'
        WHEN = 999999999.0
        SNAPSHOT_ID = 'getLastModified_simple'
        tool, ctx = self._makeOne(SNAPSHOT_ID)[1:]
        self._makeFile(tool, SNAPSHOT_ID, FILENAME, printable_bytes,
                       mod_time=WHEN)

        lm = ctx.getLastModified(FILENAME)
        self.assertIsInstance(lm, DateTime)
        self.assertEqual(lm, DateTime(WHEN))

    def test_getLastModified_subdir(self):

        FILENAME = 'subdir.txt'
        SUBDIR = 'subdir'
        PATH = f'{SUBDIR}/{FILENAME}'
        WHEN = 999999999.0
        SNAPSHOT_ID = 'getLastModified_subdir'
        tool, ctx = self._makeOne(SNAPSHOT_ID)[1:]
        self._makeFile(tool, SNAPSHOT_ID, FILENAME, printable_bytes,
                       mod_time=WHEN, subdir=SUBDIR)

        lm = ctx.getLastModified(PATH)
        self.assertIsInstance(lm, DateTime)
        self.assertEqual(lm, DateTime(WHEN))

    def test_getLastModified_directory(self):

        FILENAME = 'subdir.txt'
        SUBDIR = 'subdir'
        WHEN = 999999999.0
        SNAPSHOT_ID = 'getLastModified_directory'
        tool, ctx = self._makeOne(SNAPSHOT_ID)[1:]
        self._makeFile(tool, SNAPSHOT_ID, FILENAME, printable_bytes,
                       mod_time=WHEN, subdir=SUBDIR)

        lm = ctx.getLastModified(SUBDIR)
        self.assertIsInstance(lm, DateTime)
        self.assertEqual(lm, DateTime(WHEN))

    def test_isDirectory_nonesuch(self):

        SNAPSHOT_ID = 'isDirectory_nonesuch'
        FILENAME = 'nonesuch.txt'

        site, tool, ctx = self._makeOne(SNAPSHOT_ID)

        self.assertEqual(ctx.isDirectory(FILENAME), None)

    def test_isDirectory_simple(self):

        SNAPSHOT_ID = 'isDirectory_simple'
        FILENAME = 'simple.txt'

        site, tool, ctx = self._makeOne(SNAPSHOT_ID)
        self._makeFile(tool, SNAPSHOT_ID, FILENAME, printable_bytes)

        self.assertEqual(ctx.isDirectory(FILENAME), False)

    def test_isDirectory_nested(self):

        SNAPSHOT_ID = 'isDirectory_nested'
        SUBDIR = 'subdir'
        FILENAME = 'nested.txt'
        PATH = f'{SUBDIR}/{FILENAME}'

        site, tool, ctx = self._makeOne(SNAPSHOT_ID)
        self._makeFile(tool, SNAPSHOT_ID, FILENAME, printable_bytes,
                       subdir=SUBDIR)

        self.assertEqual(ctx.isDirectory(PATH), False)

    def test_isDirectory_subdir(self):

        SNAPSHOT_ID = 'isDirectory_subdir'
        SUBDIR = 'subdir'
        FILENAME = 'nested.txt'

        site, tool, ctx = self._makeOne(SNAPSHOT_ID)
        self._makeFile(tool, SNAPSHOT_ID, FILENAME, printable_bytes,
                       subdir=SUBDIR)

        self.assertEqual(ctx.isDirectory(SUBDIR), True)

    def test_listDirectory_nonesuch(self):

        SNAPSHOT_ID = 'listDirectory_nonesuch'
        SUBDIR = 'nonesuch/path'

        site, tool, ctx = self._makeOne(SNAPSHOT_ID)

        self.assertEqual(ctx.listDirectory(SUBDIR), None)

    def test_listDirectory_root(self):

        SNAPSHOT_ID = 'listDirectory_root'
        FILENAME = 'simple.txt'

        site, tool, ctx = self._makeOne(SNAPSHOT_ID)
        self._makeFile(tool, SNAPSHOT_ID, FILENAME, printable_bytes)

        self.assertEqual(len(ctx.listDirectory(None)), 1)
        self.assertIn(FILENAME, ctx.listDirectory(None))

    def test_listDirectory_simple(self):

        SNAPSHOT_ID = 'listDirectory_simple'
        FILENAME = 'simple.txt'

        site, tool, ctx = self._makeOne(SNAPSHOT_ID)
        self._makeFile(tool, SNAPSHOT_ID, FILENAME, printable_bytes)

        self.assertEqual(ctx.listDirectory(FILENAME), None)

    def test_listDirectory_nested(self):

        SNAPSHOT_ID = 'listDirectory_nested'
        SUBDIR = 'subdir'
        FILENAME = 'nested.txt'
        PATH = f'{SUBDIR}/{FILENAME}'

        site, tool, ctx = self._makeOne(SNAPSHOT_ID)
        self._makeFile(tool, SNAPSHOT_ID, FILENAME, printable_bytes,
                       subdir=SUBDIR)

        self.assertEqual(ctx.listDirectory(PATH), None)

    def test_listDirectory_single(self):

        SNAPSHOT_ID = 'listDirectory_nested'
        SUBDIR = 'subdir'
        FILENAME = 'nested.txt'

        site, tool, ctx = self._makeOne(SNAPSHOT_ID)
        self._makeFile(tool, SNAPSHOT_ID, FILENAME, printable_bytes,
                       subdir=SUBDIR)

        names = ctx.listDirectory(SUBDIR)
        self.assertEqual(len(names), 1)
        self.assertIn(FILENAME, names)

    def test_listDirectory_multiple(self):

        from string import ascii_uppercase

        SNAPSHOT_ID = 'listDirectory_nested'
        SUBDIR = 'subdir'
        FILENAME1 = 'nested.txt'
        FILENAME2 = 'another.txt'

        site, tool, ctx = self._makeOne(SNAPSHOT_ID)
        self._makeFile(tool, SNAPSHOT_ID, FILENAME1, printable_bytes,
                       subdir=SUBDIR)
        self._makeFile(tool, SNAPSHOT_ID, FILENAME2,
                       ascii_uppercase.encode('utf-8'), subdir=SUBDIR)

        names = ctx.listDirectory(SUBDIR)
        self.assertEqual(len(names), 2)
        self.assertIn(FILENAME1, names)
        self.assertIn(FILENAME2, names)

    def test_listDirectory_skip(self):

        from string import ascii_uppercase

        SNAPSHOT_ID = 'listDirectory_nested'
        SUBDIR = 'subdir'
        FILENAME1 = 'nested.txt'
        FILENAME2 = 'another.txt'
        FILENAME3 = 'another.bak'

        site, tool, ctx = self._makeOne(SNAPSHOT_ID)
        self._makeFile(tool, SNAPSHOT_ID, FILENAME1, printable_bytes,
                       subdir=SUBDIR)
        self._makeFile(tool, SNAPSHOT_ID, FILENAME2,
                       ascii_uppercase.encode('utf-8'), subdir=SUBDIR)
        self._makeFile(tool, SNAPSHOT_ID, FILENAME3, b'abc', subdir=SUBDIR)

        names = ctx.listDirectory(SUBDIR, skip=(FILENAME1,),
                                  skip_suffixes=('.bak',))
        self.assertEqual(len(names), 1)
        self.assertNotIn(FILENAME1, names)
        self.assertIn(FILENAME2, names)
        self.assertNotIn(FILENAME3, names)


def test_suite():
    loader = unittest.defaultTestLoader
    return unittest.TestSuite((
        loader.loadTestsFromTestCase(DirectoryImportContextTests),
        loader.loadTestsFromTestCase(DirectoryExportContextTests),
        loader.loadTestsFromTestCase(TarballImportContextTests),
        loader.loadTestsFromTestCase(TarballExportContextTests),
        loader.loadTestsFromTestCase(SnapshotExportContextTests),
        loader.loadTestsFromTestCase(SnapshotImportContextTests),
    ))



##############################################################################
#
# Copyright (c) 2004 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" GenericSetup rolemap export / import unit tests
"""

import unittest

from OFS.Folder import Folder

from ..testing import ExportImportZCMLLayer
from .common import BaseRegistryTests
from .common import DummyExportContext
from .common import DummyImportContext


class RolemapExportConfiguratorTests(BaseRegistryTests):

    layer = ExportImportZCMLLayer

    def _getTargetClass(self):

        from ..rolemap import RolemapExportConfigurator
        return RolemapExportConfigurator

    def test_listRoles_normal(self):

        EXPECTED = ['Anonymous', 'Authenticated', 'Manager', 'Owner']
        self.app.site = Folder(id='site')
        site = self.app.site
        configurator = self._makeOne(site)

        roles = list(configurator.listRoles())
        self.assertEqual(len(roles), len(EXPECTED))

        roles.sort()

        for found, expected in zip(roles, EXPECTED):
            self.assertEqual(found, expected)

    def test_listRoles_added(self):

        EXPECTED = ['Anonymous', 'Authenticated', 'Manager', 'Owner', 'ZZZ']
        self.app.site = Folder(id='site')
        site = self.app.site
        existing_roles = list(getattr(site, '__ac_roles__', []))[:]
        existing_roles.append('ZZZ')
        site.__ac_roles__ = existing_roles

        configurator = self._makeOne(site)

        roles = list(configurator.listRoles())
        self.assertEqual(len(roles), len(EXPECTED))

        roles.sort()

        for found, expected in zip(roles, EXPECTED):
            self.assertEqual(found, expected)

    def test_listPermissions_nooverrides(self):

        site = Folder(id='site').__of__(self.app)
        configurator = self._makeOne(site)

        self.assertEqual(len(configurator.listPermissions()), 0)

    def test_listPermissions_acquire(self):

        ACI = 'Access contents information'
        ROLES = ['Manager', 'Owner']

        site = Folder(id='site').__of__(self.app)
        site.manage_permission(ACI, ROLES, acquire=1)
        configurator = self._makeOne(site)

        self.assertEqual(len(configurator.listPermissions()), 1)
        info = configurator.listPermissions()[0]
        self.assertEqual(info['name'], ACI)
        self.assertEqual(info['roles'], ROLES)
        self.assertTrue(info['acquire'])

    def test_listPermissions_no_acquire(self):

        ACI = 'Access contents information'
        ROLES = ['Manager', 'Owner']

        site = Folder(id='site').__of__(self.app)
        site.manage_permission(ACI, ROLES)
        configurator = self._makeOne(site)

        self.assertEqual(len(configurator.listPermissions()), 1)
        info = configurator.listPermissions()[0]
        self.assertEqual(info['name'], ACI)
        self.assertEqual(info['roles'], ROLES)
        self.assertFalse(info['acquire'])

    def test_generateXML_empty(self):

        self.app.site = Folder(id='site')
        site = self.app.site
        configurator = self._makeOne(site).__of__(site)

        self._compareDOM(configurator.generateXML(), _EMPTY_EXPORT)

    def test_generateXML_added_role(self):

        self.app.site = Folder(id='site')
        site = self.app.site
        existing_roles = list(getattr(site, '__ac_roles__', []))[:]
        existing_roles.append('ZZZ')
        site.__ac_roles__ = existing_roles
        configurator = self._makeOne(site).__of__(site)

        self._compareDOM(configurator.generateXML(), _ADDED_ROLE_EXPORT)

    def test_generateEXML_acquired_perm(self):

        ACI = 'Access contents information'
        ROLES = ['Manager', 'Owner']

        site = Folder(id='site').__of__(self.app)
        site.manage_permission(ACI, ROLES, acquire=1)
        configurator = self._makeOne(site).__of__(site)

        self._compareDOM(configurator.generateXML(), _ACQUIRED_EXPORT)

    def test_generateEXML_unacquired_perm(self):

        ACI = 'Access contents information'
        ROLES = ['Manager', 'Owner', 'ZZZ']

        site = Folder(id='site').__of__(self.app)
        existing_roles = list(getattr(site, '__ac_roles__', []))[:]
        existing_roles.append('ZZZ')
        site.__ac_roles__ = existing_roles
        site.manage_permission(ACI, ROLES)
        configurator = self._makeOne(site).__of__(site)

        self._compareDOM(configurator.generateXML(), _COMBINED_EXPORT)

    def test_generateEXML_unacquired_perm_added_role(self):

        ACI = 'Access contents information'
        ROLES = ['Manager', 'Owner']

        site = Folder(id='site').__of__(self.app)
        site.manage_permission(ACI, ROLES)
        configurator = self._makeOne(site).__of__(site)

        self._compareDOM(configurator.generateXML(), _UNACQUIRED_EXPORT)


class RolemapImportConfiguratorTests(BaseRegistryTests):

    layer = ExportImportZCMLLayer

    def _getTargetClass(self):

        from ..rolemap import RolemapImportConfigurator
        return RolemapImportConfigurator

    def test_parseXML_empty(self):

        self.app.site = Folder(id='site')
        site = self.app.site
        existing_roles = list(getattr(site, '__ac_roles__', []))[:]
        configurator = self._makeOne(site)

        rolemap_info = configurator.parseXML(_EMPTY_EXPORT)

        self.assertEqual(len(rolemap_info['roles']), len(existing_roles))
        self.assertEqual(len(rolemap_info['permissions']), 0)

    def test_parseXML_added_role(self):

        self.app.site = Folder(id='site')
        site = self.app.site
        configurator = self._makeOne(site)

        rolemap_info = configurator.parseXML(_ADDED_ROLE_EXPORT)
        roles = rolemap_info['roles']

        self.assertEqual(len(roles), 5)
        self.assertIn('Anonymous', roles)
        self.assertIn('Authenticated', roles)
        self.assertIn('Manager', roles)
        self.assertIn('Owner', roles)
        self.assertIn('ZZZ', roles)

    def test_parseXML_acquired_permission(self):

        ACI = 'Access contents information'

        self.app.site = Folder(id='site')
        site = self.app.site
        configurator = self._makeOne(site)

        rolemap_info = configurator.parseXML(_ACQUIRED_EXPORT)

        self.assertEqual(len(rolemap_info['permissions']), 1)
        permission = rolemap_info['permissions'][0]

        self.assertEqual(permission['name'], ACI)
        self.assertTrue(permission['acquire'])

        p_roles = permission['roles']
        self.assertEqual(len(p_roles), 2)
        self.assertIn('Manager', p_roles)
        self.assertIn('Owner', p_roles)

    def test_parseXML_unacquired_permission(self):

        ACI = 'Access contents information'

        self.app.site = Folder(id='site')
        site = self.app.site
        configurator = self._makeOne(site)

        rolemap_info = configurator.parseXML(_UNACQUIRED_EXPORT)

        self.assertEqual(len(rolemap_info['permissions']), 1)
        permission = rolemap_info['permissions'][0]

        self.assertEqual(permission['name'], ACI)
        self.assertFalse(permission['acquire'])

        p_roles = permission['roles']
        self.assertEqual(len(p_roles), 2)
        self.assertIn('Manager', p_roles)
        self.assertIn('Owner', p_roles)

    def test_parseXML_unacquired_permission_added_role(self):

        ACI = 'Access contents information'

        self.app.site = Folder(id='site')
        site = self.app.site
        configurator = self._makeOne(site)

        rolemap_info = configurator.parseXML(_COMBINED_EXPORT)
        roles = rolemap_info['roles']

        self.assertEqual(len(roles), 5)
        self.assertIn('Anonymous', roles)
        self.assertIn('Authenticated', roles)
        self.assertIn('Manager', roles)
        self.assertIn('Owner', roles)
        self.assertIn('ZZZ', roles)

        self.assertEqual(len(rolemap_info['permissions']), 1)
        permission = rolemap_info['permissions'][0]

        self.assertEqual(permission['name'], ACI)
        self.assertFalse(permission['acquire'])

        p_roles = permission['roles']
        self.assertEqual(len(p_roles), 3)
        self.assertIn('Manager', p_roles)
        self.assertIn('Owner', p_roles)
        self.assertIn('ZZZ', p_roles)


_EMPTY_EXPORT = """\
<?xml version="1.0"?>
<rolemap>
  <roles>
    <role name="Anonymous"/>
    <role name="Authenticated"/>
    <role name="Manager"/>
    <role name="Owner"/>
  </roles>
  <permissions>
  </permissions>
</rolemap>
"""

_ADDED_ROLE_EXPORT = """\
<?xml version="1.0"?>
<rolemap>
  <roles>
    <role name="Anonymous"/>
    <role name="Authenticated"/>
    <role name="Manager"/>
    <role name="Owner"/>
    <role name="ZZZ"/>
  </roles>
  <permissions>
  </permissions>
</rolemap>
"""

_ACQUIRED_EXPORT = """\
<?xml version="1.0"?>
<rolemap>
  <roles>
    <role name="Anonymous"/>
    <role name="Authenticated"/>
    <role name="Manager"/>
    <role name="Owner"/>
  </roles>
  <permissions>
    <permission name="Access contents information"
                acquire="True">
      <role name="Manager"/>
      <role name="Owner"/>
    </permission>
  </permissions>
</rolemap>
"""

_UNACQUIRED_EXPORT = """\
<?xml version="1.0"?>
<rolemap>
  <roles>
    <role name="Anonymous"/>
    <role name="Authenticated"/>
    <role name="Manager"/>
    <role name="Owner"/>
  </roles>
  <permissions>
    <permission name="Access contents information"
                acquire="False">
      <role name="Manager"/>
      <role name="Owner"/>
    </permission>
  </permissions>
</rolemap>
"""

_COMBINED_EXPORT = """\
<?xml version="1.0"?>
<rolemap>
  <roles>
    <role name="Anonymous"/>
    <role name="Authenticated"/>
    <role name="Manager"/>
    <role name="Owner"/>
    <role name="ZZZ"/>
  </roles>
  <permissions>
    <permission name="Access contents information"
                acquire="False">
      <role name="Manager"/>
      <role name="Owner"/>
      <role name="ZZZ"/>
    </permission>
  </permissions>
</rolemap>
"""


class Test_exportRolemap(BaseRegistryTests):

    layer = ExportImportZCMLLayer

    def test_unchanged(self):

        self.app.site = Folder('site')
        site = self.app.site

        context = DummyExportContext(site)

        from ..rolemap import exportRolemap
        exportRolemap(context)

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'rolemap.xml')
        self._compareDOM(text.decode('utf-8'), _EMPTY_EXPORT)
        self.assertEqual(content_type, 'text/xml')

    def test_added_role(self):

        self.app.site = Folder('site')
        site = self.app.site
        existing_roles = list(getattr(site, '__ac_roles__', []))[:]
        existing_roles.append('ZZZ')
        site.__ac_roles__ = existing_roles

        context = DummyExportContext(site)

        from ..rolemap import exportRolemap
        exportRolemap(context)

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'rolemap.xml')
        self._compareDOM(text.decode('utf-8'), _ADDED_ROLE_EXPORT)
        self.assertEqual(content_type, 'text/xml')

    def test_acquired_perm(self):

        ACI = 'Access contents information'
        ROLES = ['Manager', 'Owner']

        self.app.site = Folder('site')
        site = self.app.site
        site.manage_permission(ACI, ROLES, acquire=1)

        context = DummyExportContext(site)

        from ..rolemap import exportRolemap
        exportRolemap(context)

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'rolemap.xml')
        self._compareDOM(text.decode('utf-8'), _ACQUIRED_EXPORT)
        self.assertEqual(content_type, 'text/xml')

    def test_unacquired_perm(self):

        ACI = 'Access contents information'
        ROLES = ['Manager', 'Owner', 'ZZZ']

        self.app.site = Folder('site')
        site = self.app.site
        existing_roles = list(getattr(site, '__ac_roles__', []))[:]
        existing_roles.append('ZZZ')
        site.__ac_roles__ = existing_roles
        site.manage_permission(ACI, ROLES)

        context = DummyExportContext(site)

        from ..rolemap import exportRolemap
        exportRolemap(context)

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'rolemap.xml')
        self._compareDOM(text.decode('utf-8'), _COMBINED_EXPORT)
        self.assertEqual(content_type, 'text/xml')

    def test_unacquired_perm_added_role(self):

        ACI = 'Access contents information'
        ROLES = ['Manager', 'Owner']

        self.app.site = Folder('site')
        site = self.app.site
        site.manage_permission(ACI, ROLES)

        context = DummyExportContext(site)

        from ..rolemap import exportRolemap
        exportRolemap(context)

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'rolemap.xml')
        self._compareDOM(text.decode('utf-8'), _UNACQUIRED_EXPORT)
        self.assertEqual(content_type, 'text/xml')


class Test_importRolemap(BaseRegistryTests):

    layer = ExportImportZCMLLayer

    def test_existing_roles_missing_xml_doesnt_purge(self):
        # LP # 279294
        ACI = 'Access contents information'
        VIEW = 'View'

        self.app.site = Folder(id='site')
        site = self.app.site  # wrap
        original_roles = list(getattr(site, '__ac_roles__', []))[:]
        modified_roles = original_roles[:]
        modified_roles.append('ZZZ')

        site.__ac_roles__ = modified_roles
        site.manage_permission(VIEW, ())
        site.manage_permission(ACI, ('Manager', 'ZZZ'))

        existing_allowed = [x['name'] for x in site.rolesOfPermission(ACI)
                            if x['selected']]

        self.assertEqual(existing_allowed, ['Manager', 'ZZZ'])

        context = DummyImportContext(site, True)
        # context._files[ 'rolemap.xml' ] = _EMPTY_EXPORT # no file!

        from ..rolemap import importRolemap
        importRolemap(context)

        new_roles = list(getattr(site, '__ac_roles__', []))[:]

        modified_roles.sort()
        new_roles.sort()

        self.assertEqual(modified_roles, new_roles)

        new_allowed = [x['name'] for x in site.rolesOfPermission(ACI)
                       if x['selected']]

        self.assertEqual(new_allowed, existing_allowed)

    def test_empty_default_purge(self):

        self.app.site = Folder(id='site')
        site = self.app.site
        original_roles = list(getattr(site, '__ac_roles__', []))[:]
        modified_roles = original_roles[:]
        modified_roles.append('ZZZ')
        site.__ac_roles__ = modified_roles

        context = DummyImportContext(site)
        context._files['rolemap.xml'] = _EMPTY_EXPORT

        from ..rolemap import importRolemap
        importRolemap(context)

        new_roles = list(getattr(site, '__ac_roles__', []))[:]

        original_roles.sort()
        new_roles.sort()

        self.assertEqual(original_roles, new_roles)

    def test_empty_explicit_purge(self):

        self.app.site = Folder(id='site')
        site = self.app.site
        original_roles = list(getattr(site, '__ac_roles__', []))[:]
        modified_roles = original_roles[:]
        modified_roles.append('ZZZ')
        site.__ac_roles__ = modified_roles

        context = DummyImportContext(site, True)
        context._files['rolemap.xml'] = _EMPTY_EXPORT

        from ..rolemap import importRolemap
        importRolemap(context)

        new_roles = list(getattr(site, '__ac_roles__', []))[:]

        original_roles.sort()
        new_roles.sort()

        self.assertEqual(original_roles, new_roles)

    def test_empty_skip_purge(self):

        self.app.site = Folder(id='site')
        site = self.app.site
        original_roles = list(getattr(site, '__ac_roles__', []))[:]
        modified_roles = original_roles[:]
        modified_roles.append('ZZZ')
        site.__ac_roles__ = modified_roles

        context = DummyImportContext(site, False)
        context._files['rolemap.xml'] = _EMPTY_EXPORT

        from ..rolemap import importRolemap
        importRolemap(context)

        new_roles = list(getattr(site, '__ac_roles__', []))[:]

        modified_roles.sort()
        new_roles.sort()

        self.assertEqual(modified_roles, new_roles)

    def test_acquired_permission_explicit_purge(self):

        ACI = 'Access contents information'
        VIEW = 'View'

        self.app.site = Folder(id='site')
        site = self.app.site
        site.manage_permission(ACI, ())
        site.manage_permission(VIEW, ())

        existing_allowed = [x['name']
                            for x in site.rolesOfPermission(ACI)
                            if x['selected']]

        self.assertEqual(existing_allowed, [])

        self.assertFalse(site.acquiredRolesAreUsedBy(ACI))
        self.assertFalse(site.acquiredRolesAreUsedBy(VIEW))

        context = DummyImportContext(site, True)
        context._files['rolemap.xml'] = _ACQUIRED_EXPORT

        from ..rolemap import importRolemap
        importRolemap(context)

        new_allowed = [x['name']
                       for x in site.rolesOfPermission(ACI)
                       if x['selected']]

        self.assertEqual(new_allowed, ['Manager', 'Owner'])

        # ACI is overwritten by XML, but VIEW was purged
        self.assertTrue(site.acquiredRolesAreUsedBy(ACI))
        self.assertTrue(site.acquiredRolesAreUsedBy(VIEW))

    def test_acquired_permission_no_purge(self):

        ACI = 'Access contents information'
        VIEW = 'View'

        self.app.site = Folder(id='site')
        site = self.app.site
        site.manage_permission(ACI, ())
        site.manage_permission(VIEW, ())

        existing_allowed = [x['name']
                            for x in site.rolesOfPermission(ACI)
                            if x['selected']]

        self.assertEqual(existing_allowed, [])

        self.assertFalse(site.acquiredRolesAreUsedBy(ACI))

        context = DummyImportContext(site, False)
        context._files['rolemap.xml'] = _ACQUIRED_EXPORT

        from ..rolemap import importRolemap
        importRolemap(context)

        new_allowed = [x['name']
                       for x in site.rolesOfPermission(ACI)
                       if x['selected']]

        self.assertEqual(new_allowed, ['Manager', 'Owner'])

        # ACI is overwritten by XML, but VIEW is not
        self.assertTrue(site.acquiredRolesAreUsedBy(ACI))
        self.assertFalse(site.acquiredRolesAreUsedBy(VIEW))

    def test_unacquired_permission_explicit_purge(self):

        ACI = 'Access contents information'
        VIEW = 'View'

        self.app.site = Folder(id='site')
        site = self.app.site
        site.manage_permission(VIEW, ())

        existing_allowed = [x['name']
                            for x in site.rolesOfPermission(ACI)
                            if x['selected']]

        self.assertEqual(existing_allowed, ['Manager'])

        self.assertTrue(site.acquiredRolesAreUsedBy(ACI))
        self.assertFalse(site.acquiredRolesAreUsedBy(VIEW))

        context = DummyImportContext(site, True)
        context._files['rolemap.xml'] = _UNACQUIRED_EXPORT

        from ..rolemap import importRolemap
        importRolemap(context)

        new_allowed = [x['name']
                       for x in site.rolesOfPermission(ACI)
                       if x['selected']]

        self.assertEqual(new_allowed, ['Manager', 'Owner'])

        self.assertFalse(site.acquiredRolesAreUsedBy(ACI))
        self.assertTrue(site.acquiredRolesAreUsedBy(VIEW))

    def test_unacquired_permission_skip_purge(self):

        ACI = 'Access contents information'
        VIEW = 'View'

        self.app.site = Folder(id='site')
        site = self.app.site
        site.manage_permission(VIEW, ())

        existing_allowed = [x['name']
                            for x in site.rolesOfPermission(ACI)
                            if x['selected']]

        self.assertEqual(existing_allowed, ['Manager'])

        self.assertTrue(site.acquiredRolesAreUsedBy(ACI))
        self.assertFalse(site.acquiredRolesAreUsedBy(VIEW))

        context = DummyImportContext(site, False)
        context._files['rolemap.xml'] = _UNACQUIRED_EXPORT

        from ..rolemap import importRolemap
        importRolemap(context)

        new_allowed = [x['name']
                       for x in site.rolesOfPermission(ACI)
                       if x['selected']]

        self.assertEqual(new_allowed, ['Manager', 'Owner'])

        self.assertFalse(site.acquiredRolesAreUsedBy(ACI))
        self.assertFalse(site.acquiredRolesAreUsedBy(VIEW))

    def test_unacquired_permission_added_role_explicit_purge(self):

        ACI = 'Access contents information'
        VIEW = 'View'

        self.app.site = Folder(id='site')
        site = self.app.site
        site.manage_permission(VIEW, ())

        existing_allowed = [x['name']
                            for x in site.rolesOfPermission(ACI)
                            if x['selected']]

        self.assertEqual(existing_allowed, ['Manager'])

        self.assertTrue(site.acquiredRolesAreUsedBy(ACI))
        self.assertFalse(site.acquiredRolesAreUsedBy(VIEW))

        self.assertFalse(site._has_user_defined_role('ZZZ'))

        context = DummyImportContext(site, True)
        context._files['rolemap.xml'] = _COMBINED_EXPORT

        from ..rolemap import importRolemap
        importRolemap(context)

        self.assertTrue(site._has_user_defined_role('ZZZ'))

        new_allowed = [x['name']
                       for x in site.rolesOfPermission(ACI)
                       if x['selected']]

        self.assertEqual(new_allowed, ['Manager', 'Owner', 'ZZZ'])

        self.assertFalse(site.acquiredRolesAreUsedBy(ACI))
        self.assertTrue(site.acquiredRolesAreUsedBy(VIEW))

    def test_unacquired_permission_added_role_skip_purge(self):

        ACI = 'Access contents information'
        VIEW = 'View'

        self.app.site = Folder(id='site')
        site = self.app.site
        site.manage_permission(VIEW, ())

        existing_allowed = [x['name']
                            for x in site.rolesOfPermission(ACI)
                            if x['selected']]

        self.assertEqual(existing_allowed, ['Manager'])

        self.assertTrue(site.acquiredRolesAreUsedBy(ACI))
        self.assertFalse(site.acquiredRolesAreUsedBy(VIEW))

        self.assertFalse(site._has_user_defined_role('ZZZ'))

        context = DummyImportContext(site, False)
        context._files['rolemap.xml'] = _COMBINED_EXPORT

        from ..rolemap import importRolemap
        importRolemap(context)

        self.assertTrue(site._has_user_defined_role('ZZZ'))

        new_allowed = [x['name']
                       for x in site.rolesOfPermission(ACI)
                       if x['selected']]

        self.assertEqual(new_allowed, ['Manager', 'Owner', 'ZZZ'])

        self.assertFalse(site.acquiredRolesAreUsedBy(ACI))
        self.assertFalse(site.acquiredRolesAreUsedBy(VIEW))

    def test_unacquired_permission_added_role_skip_purge_encode_ascii(self):

        ACI = 'Access contents information'
        VIEW = 'View'

        self.app.site = Folder(id='site')
        site = self.app.site
        site.manage_permission(VIEW, ())

        existing_allowed = [x['name']
                            for x in site.rolesOfPermission(ACI)
                            if x['selected']]

        self.assertEqual(existing_allowed, ['Manager'])

        self.assertTrue(site.acquiredRolesAreUsedBy(ACI))
        self.assertFalse(site.acquiredRolesAreUsedBy(VIEW))

        self.assertFalse(site._has_user_defined_role('ZZZ'))

        context = DummyImportContext(site, False, encoding='ascii')
        context._files['rolemap.xml'] = _COMBINED_EXPORT

        from ..rolemap import importRolemap
        importRolemap(context)

        self.assertTrue(site._has_user_defined_role('ZZZ'))

        new_allowed = [x['name']
                       for x in site.rolesOfPermission(ACI)
                       if x['selected']]

        self.assertEqual(new_allowed, ['Manager', 'Owner', 'ZZZ'])

        self.assertFalse(site.acquiredRolesAreUsedBy(ACI))
        self.assertFalse(site.acquiredRolesAreUsedBy(VIEW))


def test_suite():
    loader = unittest.defaultTestLoader
    return unittest.TestSuite((
        loader.loadTestsFromTestCase(RolemapExportConfiguratorTests),
        loader.loadTestsFromTestCase(RolemapImportConfiguratorTests),
        loader.loadTestsFromTestCase(Test_exportRolemap),
        loader.loadTestsFromTestCase(Test_importRolemap),
    ))



##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Filesystem exporter / importer adapter unit tests.
"""

import unittest
from io import StringIO

from .conformance import ConformsToIFilesystemExporter
from .conformance import ConformsToIFilesystemImporter
from .conformance import ConformsToIINIAware


class SimpleINIAwareTests(unittest.TestCase, ConformsToIINIAware):

    def _getTargetClass(self):
        from ..content import SimpleINIAware
        return SimpleINIAware

    def test_as_ini_no_properties(self):
        context = _makePropertied('no_properties')
        context._properties = ()
        adapter = self._getTargetClass()(context)
        text = adapter.as_ini()
        parser = _parseINI(text)
        self.assertFalse(parser.sections())
        default_options = parser.defaults()
        self.assertEqual(len(default_options), 0)

    def test_as_ini_string_property(self):
        TITLE = 'String Property'
        DESCR = 'Another property'
        context = _makePropertied('string_property')
        context.title = TITLE
        context._setProperty('description', DESCR)
        adapter = self._getTargetClass()(context)
        text = adapter.as_ini()
        parser = _parseINI(text)
        self.assertFalse(parser.sections())
        default_options = parser.defaults()
        self.assertEqual(len(default_options), 2)
        self.assertEqual(default_options['title'].strip(), TITLE)
        self.assertEqual(default_options['description'].strip(), DESCR)

    def test_as_ini_other_properties(self):
        from DateTime.DateTime import DateTime
        INTPROP = 42
        FLOATPROP = 3.1415926
        DATESTR = '2005-11-07T12:00:00.000Z'
        context = _makePropertied('string_property')
        context._properties = ()
        context._setProperty('int_prop', INTPROP, 'int')
        context._setProperty('float_prop', FLOATPROP, 'float')
        context._setProperty('date_prop', DateTime(DATESTR), 'date')
        adapter = self._getTargetClass()(context)
        text = adapter.as_ini()
        parser = _parseINI(text)
        self.assertFalse(parser.sections())
        default_options = parser.defaults()
        self.assertEqual(len(default_options), 3)
        self.assertEqual(default_options['int_prop'], str(INTPROP))
        self.assertEqual(default_options['float_prop'], str(FLOATPROP))
        self.assertEqual(default_options['date_prop'], str(DateTime(DATESTR)))

    def test_put_ini_empty(self):
        context = _makePropertied('empty_ini')
        adapter = self._getTargetClass()(context)
        context._properties = ()
        self.assertFalse(context.propertyItems())
        adapter.put_ini('')
        self.assertFalse(context.propertyItems())

    def test_put_ini_with_values_stripped(self):
        context = _makePropertied('empty_ini')
        adapter = self._getTargetClass()(context)
        adapter.put_ini('[DEFAULT]\ntitle = Foo \ndescription = bar ')
        props = context.propdict()
        self.assertEqual(len(props), 2)
        self.assertIn('title', props)
        self.assertIn('description', props)
        self.assertEqual(context.title, 'Foo')
        self.assertEqual(context.description, 'bar')

    def test_put_ini_other_properties(self):
        from DateTime.DateTime import DateTime
        INTPROP = 42
        FLOATPROP = 3.1415926
        DATESTR = '2005-11-07T12:00:00.000Z'
        DATESTR2 = '2005-11-09T12:00:00.000Z'
        context = _makePropertied('string_property')
        context._properties = ()
        context._setProperty('int_prop', INTPROP, 'int')
        context._setProperty('float_prop', FLOATPROP, 'float')
        context._setProperty('date_prop', DateTime(DATESTR), 'date')
        adapter = self._getTargetClass()(context)
        adapter.put_ini('''\
[DEFAULT]
int_prop = 13
\nfloat_prop = 2.818
\ndate_prop = %s''' % DATESTR2)
        self.assertEqual(len(context.propertyIds()), 3)
        self.assertEqual(context.int_prop, 13)
        self.assertEqual(context.float_prop, 2.818)
        self.assertEqual(context.date_prop, DateTime(DATESTR2))


class FolderishExporterImporterTests(unittest.TestCase):

    def setUp(self):
        from zope.component.testing import setUp
        setUp()

    def tearDown(self):
        from zope.component.testing import tearDown
        tearDown()

    def _getExporter(self):
        from ..content import exportSiteStructure
        return exportSiteStructure

    def _getImporter(self):
        from ..content import importSiteStructure
        return importSiteStructure

    def _makeSetupTool(self):
        from ..tool import SetupTool
        return SetupTool('portal_setup')

    def _setUpAdapters(self):
        from OFS.interfaces import IObjectManager
        from OFS.interfaces import IPropertyManager
        from zope.component import getSiteManager

        from ..content import CSVAwareFileAdapter
        from ..content import DAVAwareFileAdapter
        from ..content import FolderishExporterImporter
        from ..content import INIAwareFileAdapter
        from ..content import SimpleINIAware
        from ..interfaces import ICSVAware
        from ..interfaces import IDAVAware
        from ..interfaces import IFilesystemExporter
        from ..interfaces import IFilesystemImporter
        from ..interfaces import IINIAware

        sm = getSiteManager()

        sm.registerAdapter(FolderishExporterImporter,
                           (IObjectManager,),
                           IFilesystemExporter,
                           )

        sm.registerAdapter(FolderishExporterImporter,
                           (IObjectManager,),
                           IFilesystemImporter,
                           )

        sm.registerAdapter(SimpleINIAware,
                           (IPropertyManager,),
                           IINIAware,
                           )

        sm.registerAdapter(CSVAwareFileAdapter,
                           (ICSVAware,),
                           IFilesystemExporter,
                           )

        sm.registerAdapter(CSVAwareFileAdapter,
                           (ICSVAware,),
                           IFilesystemImporter,
                           )

        sm.registerAdapter(INIAwareFileAdapter,
                           (IINIAware,),
                           IFilesystemExporter,
                           )

        sm.registerAdapter(INIAwareFileAdapter,
                           (IINIAware,),
                           IFilesystemImporter,
                           )

        sm.registerAdapter(DAVAwareFileAdapter,
                           (IDAVAware,),
                           IFilesystemExporter,
                           )

        sm.registerAdapter(DAVAwareFileAdapter,
                           (IDAVAware,),
                           IFilesystemImporter,
                           )

    def test_export_empty_site(self):
        from .common import DummyExportContext
        self._setUpAdapters()
        site = _makeFolder('site')
        site.title = 'test_export_empty_site'
        site.description = 'Testing export of an empty site.'
        context = DummyExportContext(site)
        exporter = self._getExporter()
        exporter(context)

        self.assertEqual(len(context._wrote), 2)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'structure/.objects')
        self.assertEqual(content_type, 'text/comma-separated-values')

        objects = _parseCSV(text)
        self.assertEqual(len(objects), 0)

        filename, text, content_type = context._wrote[1]
        self.assertEqual(filename, 'structure/.properties')
        self.assertEqual(content_type, 'text/plain')

        parser = _parseINI(text)

        defaults = parser.defaults()
        self.assertEqual(len(defaults), 1)
        self.assertEqual(defaults['title'], site.title)

    def test_export_empty_site_with_setup_tool(self):
        from .common import DummyExportContext
        self._setUpAdapters()
        site = _makeFolder('site')
        site._setObject('setup_tool', self._makeSetupTool())
        site._updateProperty('title', 'test_export_empty_site_with_setup_tool')
        site._setProperty('description',
                          'Testing export of an empty site with setup tool.')
        context = DummyExportContext(site)
        exporter = self._getExporter()
        exporter(context)

        self.assertEqual(len(context._wrote), 2)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'structure/.objects')
        self.assertEqual(content_type, 'text/comma-separated-values')

        objects = _parseCSV(text)
        self.assertEqual(len(objects), 0)

        filename, text, content_type = context._wrote[1]
        self.assertEqual(filename, 'structure/.properties')
        self.assertEqual(content_type, 'text/plain')

        parser = _parseINI(text)

        defaults = parser.defaults()
        self.assertEqual(len(defaults), 2)
        self.assertEqual(defaults['title'], site.title)
        self.assertEqual(defaults['description'], site.description)

    def test_export_site_with_non_exportable_simple_items(self):
        from ..utils import _getDottedName
        from .common import DummyExportContext
        self._setUpAdapters()
        ITEM_IDS = ('foo', 'bar', 'baz')

        site = _makeFolder('site')
        site.title = 'AAA'
        site._setProperty('description', 'CCC')
        item = _makeItem('aside')
        dotted = _getDottedName(item.__class__)
        for id in ITEM_IDS:
            site._setObject(id, _makeItem(id))

        context = DummyExportContext(site)
        exporter = self._getExporter()
        exporter(context)

        self.assertEqual(len(context._wrote), 2)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'structure/.objects')
        self.assertEqual(content_type, 'text/comma-separated-values')

        objects = _parseCSV(text)
        self.assertEqual(len(objects), 3)
        for index in range(len(ITEM_IDS)):
            self.assertEqual(objects[index][0], ITEM_IDS[index])
            self.assertEqual(objects[index][1], dotted)

        filename, text, content_type = context._wrote[1]
        self.assertEqual(filename, 'structure/.properties')
        self.assertEqual(content_type, 'text/plain')
        parser = _parseINI(text)

        defaults = parser.defaults()
        self.assertEqual(len(defaults), 2)
        self.assertEqual(defaults['title'], 'AAA')
        self.assertEqual(defaults['description'], 'CCC')

    def test_export_site_with_exportable_simple_items(self):
        from ..utils import _getDottedName
        from .common import DummyExportContext
        self._setUpAdapters()
        ITEM_IDS = ('foo', 'bar', 'baz')

        site = _makeFolder('site')
        site.title = 'AAA'
        site._setProperty('description', 'CCC')
        aware = _makeINIAware('aside')
        dotted = _getDottedName(aware.__class__)
        for id in ITEM_IDS:
            site._setObject(id, _makeINIAware(id))

        context = DummyExportContext(site)
        exporter = self._getExporter()
        exporter(context)

        self.assertEqual(len(context._wrote), 2 + len(ITEM_IDS))
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'structure/.objects')
        self.assertEqual(content_type, 'text/comma-separated-values')

        objects = _parseCSV(text)
        self.assertEqual(len(objects), 3)
        for index in range(len(ITEM_IDS)):
            self.assertEqual(objects[index][0], ITEM_IDS[index])
            self.assertEqual(objects[index][1], dotted)

            filename, text, content_type = context._wrote[index + 2]
            self.assertEqual(filename, 'structure/%s.ini' % ITEM_IDS[index])
            object = site._getOb(ITEM_IDS[index])
            self.assertEqual(text.strip(),
                             object.as_ini().strip())
            self.assertEqual(content_type, 'text/plain')

        filename, text, content_type = context._wrote[1]
        self.assertEqual(filename, 'structure/.properties')
        self.assertEqual(content_type, 'text/plain')
        parser = _parseINI(text)

        defaults = parser.defaults()
        self.assertEqual(len(defaults), 2)
        self.assertEqual(defaults['title'], 'AAA')
        self.assertEqual(defaults['description'], 'CCC')

    def test_export_site_with_subfolders(self):
        from ..utils import _getDottedName
        from .common import DummyExportContext
        self._setUpAdapters()
        FOLDER_IDS = ('foo', 'bar', 'baz')

        site = _makeFolder('site')
        site.title = 'AAA'
        site._setProperty('description', 'CCC')
        aside = _makeFolder('aside')
        dotted = _getDottedName(aside.__class__)
        for id in FOLDER_IDS:
            folder = _makeFolder(id)
            folder.title = 'Title: %s' % id
            site._setObject(id, folder)

        context = DummyExportContext(site)
        exporter = self._getExporter()
        exporter(context)

        self.assertEqual(len(context._wrote), 2 + (2 * len(FOLDER_IDS)))
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'structure/.objects')
        self.assertEqual(content_type, 'text/comma-separated-values')

        objects = _parseCSV(text)
        self.assertEqual(len(objects), 3)

        for index in range(len(FOLDER_IDS)):
            id = FOLDER_IDS[index]
            self.assertEqual(objects[index][0], id)
            self.assertEqual(objects[index][1], dotted)

            filename, text, content_type = context._wrote[2 + (2 * index)]
            self.assertEqual(filename, '/'.join(('structure', id, '.objects')))
            self.assertEqual(content_type, 'text/comma-separated-values')
            subobjects = _parseCSV(text)
            self.assertEqual(len(subobjects), 0)

            filename, text, content_type = context._wrote[2 + (2 * index) + 1]
            self.assertEqual(filename,
                             '/'.join(('structure', id, '.properties')))
            self.assertEqual(content_type, 'text/plain')
            parser = _parseINI(text)

            defaults = parser.defaults()
            self.assertEqual(len(defaults), 1)
            self.assertEqual(defaults['title'], 'Title: %s' % id)

        filename, text, content_type = context._wrote[1]
        self.assertEqual(filename, 'structure/.properties')
        self.assertEqual(content_type, 'text/plain')

        parser = _parseINI(text)

        defaults = parser.defaults()
        self.assertEqual(len(defaults), 2)
        self.assertEqual(defaults['title'], 'AAA')
        self.assertEqual(defaults['description'], 'CCC')

    def test_export_site_with_csvaware(self):
        from ..utils import _getDottedName
        from .common import DummyExportContext
        self._setUpAdapters()

        site = _makeFolder('site')
        site.title = 'test_export_site_with_csvaware'
        site._setProperty('description',
                          'Testing export of an site with CSV-aware content.')

        aware = _makeCSVAware('aware')
        site._setObject('aware', aware)

        context = DummyExportContext(site)
        exporter = self._getExporter()
        exporter(context)

        self.assertEqual(len(context._wrote), 3)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'structure/.objects')
        self.assertEqual(content_type, 'text/comma-separated-values')

        objects = _parseCSV(text)
        self.assertEqual(len(objects), 1)
        self.assertEqual(objects[0][0], 'aware')
        self.assertEqual(objects[0][1], _getDottedName(aware.__class__))

        filename, text, content_type = context._wrote[1]
        self.assertEqual(filename, 'structure/.properties')
        self.assertEqual(content_type, 'text/plain')

        parser = _parseINI(text)

        defaults = parser.defaults()
        self.assertEqual(len(defaults), 2)
        self.assertEqual(defaults['title'], site.title)
        self.assertEqual(defaults['description'], site.description)

        filename, text, content_type = context._wrote[2]
        self.assertEqual(filename, 'structure/aware.csv')
        self.assertEqual(content_type, 'text/comma-separated-values')
        rows = _parseCSV(text)
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0][0], 'one')
        self.assertEqual(rows[0][1], 'two')
        self.assertEqual(rows[0][2], 'three')
        self.assertEqual(rows[1][0], 'four')
        self.assertEqual(rows[1][1], 'five')
        self.assertEqual(rows[1][2], 'six')

    def test_import_empty_site(self):
        from .common import DummyImportContext
        self._setUpAdapters()
        site = _makeFolder('site')
        context = DummyImportContext(site)
        context._files['structure/.objects'] = ''
        importer = self._getImporter()
        self.assertEqual(len(site.objectIds()), 0)
        importer(context)
        self.assertEqual(len(site.objectIds()), 0)

    def test_import_empty_site_with_setup_tool(self):
        from .common import DummyImportContext
        self._setUpAdapters()
        site = _makeFolder('site')
        site._setObject('setup_tool', self._makeSetupTool())
        context = DummyImportContext(site)
        importer = self._getImporter()

        self.assertEqual(len(site.objectIds()), 1)
        self.assertEqual(site.objectIds()[0], 'setup_tool')
        importer(context)
        self.assertEqual(len(site.objectIds()), 1)
        self.assertEqual(site.objectIds()[0], 'setup_tool')

    def test_import_site_with_subfolders(self):
        from ..utils import _getDottedName
        from .common import DummyImportContext
        self._setUpAdapters()
        FOLDER_IDS = ('foo', 'bar', 'baz')

        site = _makeFolder('site')
        dotted = _getDottedName(site.__class__)

        context = DummyImportContext(site)

        for id in FOLDER_IDS:
            context._files['structure/%s/.objects' % id] = ''
            context._files['structure/%s/.properties' % id] = (
                _PROPERTIES_TEMPLATE % id)

        _ROOT_OBJECTS = '\n'.join([f'{id},{dotted}' for id in FOLDER_IDS])

        context._files['structure/.objects'] = _ROOT_OBJECTS
        context._files['structure/.properties'] = (
            _PROPERTIES_TEMPLATE % 'Test Site')

        importer = self._getImporter()
        importer(context)

        content = site.objectValues()
        self.assertEqual(len(content), len(FOLDER_IDS))

    def test_import_site_with_subitems(self):
        from ..utils import _getDottedName
        from .common import DummyImportContext
        from .faux_objects import KNOWN_INI
        from .faux_objects import TestINIAware
        dotted = _getDottedName(TestINIAware)
        self._setUpAdapters()
        ITEM_IDS = ('foo', 'bar', 'baz')

        site = _makeFolder('site')

        context = DummyImportContext(site)
        # We want to add 'baz' to 'foo', without losing 'bar'
        context._files['structure/.objects'] = '\n'.join(
            [f'{x},{dotted}' for x in ITEM_IDS])
        for index in range(len(ITEM_IDS)):
            id = ITEM_IDS[index]
            context._files[
                'structure/%s.ini' % id] = KNOWN_INI % ('Title: %s' % id,
                                                        'xyzzy',
                                                        )
        importer = self._getImporter()
        importer(context)

        after = site.objectIds()
        self.assertEqual(len(after), len(ITEM_IDS))
        for found_id, expected_id in zip(after, ITEM_IDS):
            self.assertEqual(found_id, expected_id)

    def test_import_site_with_subitems_wo_adapter(self):
        from ..utils import _getDottedName
        from .common import DummyImportContext
        item = _makeItem('no_adapter')
        dotted = _getDottedName(item.__class__)
        self._setUpAdapters()

        site = _makeFolder('site')

        context = DummyImportContext(site)
        # We want to add 'baz' to 'foo', without losing 'bar'
        context._files['structure/.objects'] = '\n'.join(
            [f'{x},{dotted}' for x in ('no_adapter',)])
        importer = self._getImporter()
        importer(context)

        after = site.objectIds()
        self.assertEqual(len(after), 1)
        self.assertEqual(after[0], 'no_adapter')

    def test_import_site_with_subitems_and_blanklines_dotobjects(self):
        from ..utils import _getDottedName
        from .common import DummyImportContext
        from .faux_objects import KNOWN_INI
        from .faux_objects import TestINIAware
        dotted = _getDottedName(TestINIAware)
        self._setUpAdapters()
        ITEM_IDS = ('foo', 'bar', 'baz')

        site = _makeFolder('site')

        context = DummyImportContext(site)
        # We want to add 'baz' to 'foo', without losing 'bar'
        correct = '\n'.join([f'{x},{dotted}' for x in ITEM_IDS])
        broken = correct + '\n\n'
        context._files['structure/.objects'] = broken
        for index in range(len(ITEM_IDS)):
            id = ITEM_IDS[index]
            context._files[
                'structure/%s.ini' % id] = KNOWN_INI % ('Title: %s' % id,
                                                        'xyzzy',
                                                        )
        importer = self._getImporter()
        importer(context)

        after = site.objectIds()
        self.assertEqual(len(after), len(ITEM_IDS))
        for found_id, expected_id in zip(after, ITEM_IDS):
            self.assertEqual(found_id, expected_id)

    def test_import_site_with_subitem_unknown_portal_type(self):
        from .common import DummyImportContext
        from .faux_objects import KNOWN_INI
        self._setUpAdapters()
        ITEM_IDS = ('foo', 'bar', 'baz')

        site = _makeFolder('site')

        context = DummyImportContext(site)
        # We want to add 'baz' to 'foo', without losing 'bar'
        context._files['structure/.objects'] = '\n'.join(
            ['%s,Unknown Type' % x for x in ITEM_IDS])
        for index in range(len(ITEM_IDS)):
            id = ITEM_IDS[index]
            context._files[
                'structure/%s.ini' % id] = KNOWN_INI % ('Title: %s' % id,
                                                        'xyzzy',
                                                        )

        importer = self._getImporter()
        importer(context)

        after = site.objectIds()
        self.assertEqual(len(after), 0)
        self.assertEqual(len(context._notes), len(ITEM_IDS))
        for level, component, message in context._notes:
            self.assertEqual(component, 'SFWA')
            self.assertTrue(message.startswith("Couldn't make"))

    def test_import_site_with_subitems_and_no_preserve(self):
        from .common import DummyImportContext
        self._setUpAdapters()
        ITEM_IDS = ('foo', 'bar', 'baz')

        site = _makeFolder('site')
        for id in ITEM_IDS:
            site._setObject(id, _makeItem(id))

        context = DummyImportContext(site)
        # We want to add 'baz' to 'foo', without losing 'bar'
        context._files['structure/.objects'] = ''

        importer = self._getImporter()
        importer(context)

        self.assertEqual(len(site.objectIds()), 0)

    def test_import_site_with_subitemss_and_preserve(self):
        from .common import DummyImportContext
        self._setUpAdapters()
        ITEM_IDS = ('foo', 'bar', 'baz')

        site = _makeFolder('site')
        for id in ITEM_IDS:
            site._setObject(id, _makeItem(id))

        context = DummyImportContext(site)
        # We want to add 'baz' to 'foo', without losing 'bar'
        context._files['structure/.objects'] = ''
        context._files['structure/.preserve'] = '*'

        importer = self._getImporter()
        importer(context)

        after = site.objectIds()
        self.assertEqual(len(after), len(ITEM_IDS))
        for i in range(len(ITEM_IDS)):
            self.assertEqual(after[i], ITEM_IDS[i])

    def test_import_site_with_subitemss_and_preserve_partial(self):
        from .common import DummyImportContext
        self._setUpAdapters()
        ITEM_IDS = ('foo', 'bar', 'baz')

        site = _makeFolder('site')
        for id in ITEM_IDS:
            site._setObject(id, _makeItem(id))

        context = DummyImportContext(site)
        # We want to add 'baz' to 'foo', without losing 'bar'
        context._files['structure/.objects'] = ''
        context._files['structure/.preserve'] = 'b*'

        importer = self._getImporter()
        importer(context)

        after = site.objectIds()
        self.assertEqual(len(after), 2)
        self.assertEqual(after[0], 'bar')
        self.assertEqual(after[1], 'baz')

    def test_import_site_with_subfolders_and_preserve(self):
        from ..utils import _getDottedName
        from .common import DummyImportContext
        self._setUpAdapters()

        site = _makeFolder('site')
        site._setObject('foo', _makeFolder('foo'))
        foo = site._getOb('foo')
        foo._setObject('bar', _makeFolder('bar'))
        bar = foo._getOb('bar')

        context = DummyImportContext(site)
        # We want to add 'baz' to 'foo', without losing 'bar'
        context._files['structure/.objects'
                       ] = 'foo,%s' % _getDottedName(foo.__class__)
        context._files['structure/.preserve'] = '*'
        context._files['structure/foo/.objects'
                       ] = 'baz,%s' % _getDottedName(bar.__class__)
        context._files['structure/foo/.preserve'] = '*'
        context._files['structure/foo/baz/.objects'] = ''

        importer = self._getImporter()
        importer(context)

        self.assertEqual(len(site.objectIds()), 1)
        self.assertEqual(site.objectIds()[0], 'foo')

        self.assertEqual(len(foo.objectIds()), 2, site.foo.objectIds())
        self.assertEqual(foo.objectIds()[0], 'bar')
        self.assertEqual(foo.objectIds()[1], 'baz')


class Test_globpattern(unittest.TestCase):

    NAMELIST = ('foo', 'bar', 'baz', 'bam', 'qux', 'quxx', 'quxxx')

    def _checkResults(self, globpattern, namelist, expected):
        from ..content import _globtest
        found = _globtest(globpattern, namelist)
        self.assertEqual(len(found), len(expected))
        for found_item, expected_item in zip(found, expected):
            self.assertEqual(found_item, expected_item)

    def test_star(self):
        self._checkResults('*', self.NAMELIST, self.NAMELIST)

    def test_simple(self):
        self._checkResults('b*', self.NAMELIST,
                           [x for x in self.NAMELIST if x.startswith('b')])

    def test_multiple(self):
        self._checkResults('b*\n*x', self.NAMELIST,
                           [x for x in self.NAMELIST
                            if x.startswith('b') or x.endswith('x')])


class CSVAwareFileAdapterTests(unittest.TestCase,
                               ConformsToIFilesystemExporter,
                               ConformsToIFilesystemImporter,
                               ):

    def _getTargetClass(self):
        from ..content import CSVAwareFileAdapter
        return CSVAwareFileAdapter

    def _makeOne(self, context, *args, **kw):
        return self._getTargetClass()(context, *args, **kw)

    def test_export_with_known_CSV(self):
        from .common import DummyExportContext
        from .faux_objects import KNOWN_CSV
        sheet = _makeCSVAware('config')

        adapter = self._makeOne(sheet)
        context = DummyExportContext(None)
        adapter.export(context, 'subpath/to/sheet')

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'subpath/to/sheet/config.csv')
        self.assertEqual(content_type, 'text/comma-separated-values')

        self.assertEqual(text.strip(), KNOWN_CSV.strip())

    def test_import_with_known_CSV(self):
        from .common import DummyImportContext
        ORIG_CSV = """\
one,two,three
four,five,six
"""
        NEW_CSV = b"""\
four,five,six
one,two,three
"""
        sheet = _makeCSVAware('config', ORIG_CSV)

        adapter = self._makeOne(sheet)
        context = DummyImportContext(None)
        context._files['subpath/to/sheet/config.csv'] = NEW_CSV
        adapter.import_(context, 'subpath/to/sheet')

        self.assertEqual(sheet._was_put.getvalue().strip(), NEW_CSV.strip())


_PROPERTIES_TEMPLATE = """
[DEFAULT]
Title = %s
Description = This is a test
"""


class INIAwareFileAdapterTests(unittest.TestCase,
                               ConformsToIFilesystemExporter,
                               ConformsToIFilesystemImporter,
                               ):

    def _getTargetClass(self):
        from ..content import INIAwareFileAdapter
        return INIAwareFileAdapter

    def _makeOne(self, context, *args, **kw):
        return self._getTargetClass()(context, *args, **kw)

    def test_export_ini_file(self):
        from .common import DummyExportContext
        ini_file = _makeINIAware('ini_file.html')
        adapter = self._makeOne(ini_file)
        context = DummyExportContext(None)
        adapter.export(context, 'subpath/to')

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'subpath/to/ini_file.html.ini')
        self.assertEqual(content_type, 'text/plain')

        self.assertEqual(text.strip(), ini_file.as_ini().strip())

    def test_import_ini_file(self):
        from .common import DummyImportContext
        from .faux_objects import KNOWN_INI
        ini_file = _makeINIAware('ini_file.html')
        adapter = self._makeOne(ini_file)
        context = DummyImportContext(None)
        context._files['subpath/to/ini_file.html.ini'] = (
            KNOWN_INI % ('Title: ini_file', 'abc'))

        adapter.import_(context, 'subpath/to')
        text = ini_file._was_put
        parser = _parseINI(text)
        self.assertEqual(parser.get('DEFAULT', 'title'), 'Title: ini_file')
        self.assertEqual(parser.get('DEFAULT', 'description'), 'abc')


class DAVAwareFileAdapterTests(unittest.TestCase,
                               ConformsToIFilesystemExporter,
                               ConformsToIFilesystemImporter,
                               ):

    def _getTargetClass(self):
        from ..content import DAVAwareFileAdapter
        return DAVAwareFileAdapter

    def _makeOne(self, context, *args, **kw):
        return self._getTargetClass()(context, *args, **kw)

    def test_export_dav_file(self):
        from .common import DummyExportContext
        dav_file = _makeDAVAware('dav_file.html')
        adapter = self._makeOne(dav_file)
        context = DummyExportContext(None)
        adapter.export(context, 'subpath/to')

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'subpath/to/dav_file.html')
        self.assertEqual(content_type, 'text/plain')
        self.assertEqual(text.strip(), dav_file.manage_FTPget().strip())

    def test_import_dav_file(self):
        from .common import DummyImportContext
        from .faux_objects import KNOWN_DAV
        VALUES = (b'Title: dav_file', b'Description: abc', b'body goes here')
        dav_file = _makeDAVAware('dav_file.html')
        adapter = self._makeOne(dav_file)
        context = DummyImportContext(None)
        context._files['subpath/to/dav_file.html'] = KNOWN_DAV % VALUES

        adapter.import_(context, 'subpath/to')
        self.assertEqual(dav_file._was_put, KNOWN_DAV % VALUES)


def _makePropertied(id):
    from .faux_objects import TestSimpleItemWithProperties

    propertied = TestSimpleItemWithProperties()
    propertied._setId(id)

    return propertied


def _makeCSVAware(id, csv=None):
    from .faux_objects import TestCSVAware

    aware = TestCSVAware()
    aware._setId(id)
    if csv is not None:
        aware._csv = csv

    return aware


def _makeINIAware(id):
    from .faux_objects import TestINIAware

    aware = TestINIAware()
    aware._setId(id)

    return aware


def _makeDAVAware(id):
    from .faux_objects import TestDAVAware

    aware = TestDAVAware()
    aware._setId(id)

    return aware


def _makeItem(id):
    from .faux_objects import TestSimpleItem

    aware = TestSimpleItem()
    aware._setId(id)

    return aware


def _makeFolder(id):
    from OFS.Folder import Folder
    from OFS.interfaces import IObjectManager
    from OFS.interfaces import IPropertyManager
    from zope.interface import directlyProvides
    from zope.interface import providedBy

    folder = Folder(id)
    directlyProvides(folder, providedBy(folder)
                     + IObjectManager + IPropertyManager)

    return folder


def _parseCSV(text):
    from csv import reader

    return [x for x in reader(StringIO(text))]


def _parseINI(text):
    from configparser import ConfigParser
    parser = ConfigParser()

    # read_file/readfp expect text, not bytes
    if isinstance(text, bytes):
        text = text.decode('UTF-8')

    parser.read_file(StringIO(text))
    return parser


def test_suite():
    loader = unittest.defaultTestLoader
    suite = unittest.TestSuite()
    suite.addTest(loader.loadTestsFromTestCase(SimpleINIAwareTests))
    suite.addTest(loader.loadTestsFromTestCase(FolderishExporterImporterTests))
    suite.addTest(loader.loadTestsFromTestCase(Test_globpattern))
    suite.addTest(loader.loadTestsFromTestCase(CSVAwareFileAdapterTests))
    suite.addTest(loader.loadTestsFromTestCase(INIAwareFileAdapterTests))
    suite.addTest(loader.loadTestsFromTestCase(DAVAwareFileAdapterTests))
    return suite



##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors>
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Upgrade steps and registry.
"""
import packaging.version
from BTrees.OOBTree import OOBTree

from Products.GenericSetup.interfaces import IUpgradeSteps
from Products.GenericSetup.registry import GlobalRegistryStorage
from Products.GenericSetup.utils import _getHash


def normalize_version(version):
    if isinstance(version, tuple):
        version = '.'.join(version)
    elif not version or version == 'unknown':
        version = '0'
    try:
        return packaging.version.parse(version)
    except packaging.version.InvalidVersion:
        # Older setuptools versions created a LegacyVersion when parsing to a
        # proper number fails.  Let's use local version segments to have a
        # strict version and still have a sort order.  See
        # https://github.com/zopefoundation/Products.GenericSetup/issues/126
        return packaging.version.parse(f'0+{version}')


def _version_matches_all(version):
    if isinstance(version, tuple):
        version = '.'.join(version)
    return version in (None, 'unknown', 'all')


def _version_matches(source, step_source, step_dest, strict=False, dest=None):
    # Step source and destination must match.
    source_matches_all = _version_matches_all(source)
    dest_matches_all = _version_matches_all(dest)
    if source_matches_all and dest_matches_all:
        return True
    source = normalize_version(source)
    if not source_matches_all:
        # Check step source.
        if not _version_matches_all(step_source):
            start = normalize_version(step_source)
            if strict:
                if start != source:
                    return False
            elif start < source:
                return False
    # Step source is okay. Now check step destination.
    if _version_matches_all(step_dest):
        return True
    stop = normalize_version(step_dest)
    # Check maximum or strictly wanted destination.
    if not dest_matches_all:
        dest = normalize_version(dest)
        if strict:
            if stop != dest:
                return False
        elif stop > dest:
            return False

    return stop > source


class UpgradeRegistry:
    """Registry of upgrade steps, by profile.

    Registry keys are profile ids.

    Each registry value is a nested mapping:
      - id -> step for single steps
      - id -> [ (id1, step1), (id2, step2) ] for nested steps
    """

    def __init__(self):
        self._registry = GlobalRegistryStorage(IUpgradeSteps)

    def __getitem__(self, key):
        return self._registry.get(key)

    def keys(self):
        return self._registry.keys()

    def clear(self):
        self._registry.clear()

    def getUpgradeStepsForProfile(self, profile_id):
        """Return the upgrade steps mapping for a given profile, or
        None if there are no steps registered for a profile matching
        that id.
        """
        prefix = 'profile-'
        if profile_id.startswith(prefix):
            profile_id = profile_id[len(prefix):]
        profile_steps = self._registry.get(profile_id)
        if profile_steps is None:
            self._registry[profile_id] = OOBTree()
            profile_steps = self._registry.get(profile_id)
        return profile_steps

    def getUpgradeStep(self, profile_id, step_id):
        """Returns the specified upgrade step for the specified
        profile, or None if it doesn't exist.
        """
        prefix = 'profile-'
        if profile_id.startswith(prefix):
            profile_id = profile_id[len(prefix):]
        profile_steps = self._registry.get(profile_id)
        if profile_steps is not None:
            step = profile_steps.get(step_id, None)
            if step is None:
                for key in profile_steps.keys():
                    if isinstance(profile_steps[key], list):
                        subs = dict(profile_steps[key])
                        step = subs.get(step_id, None)
                        if step is not None:
                            break
            elif isinstance(step, list):
                subs = dict(step)
                step = subs.get(step_id, None)
            return step


_upgrade_registry = UpgradeRegistry()


class UpgradeEntity:
    """
    Base class for actions to be taken during an upgrade process.
    """

    def __init__(self, title, profile, source, dest, desc, checker=None,
                 sortkey=0):
        self.id = _getHash(title, source, dest, sortkey)
        self.title = title
        if source == '*':
            source = None
        elif isinstance(source, str):
            source = tuple(source.split('.'))
        self.source = source
        if dest == '*':
            dest = None
        elif isinstance(dest, str):
            dest = tuple(dest.split('.'))
        self.dest = dest
        self.description = desc
        self.checker = checker
        self.sortkey = sortkey
        self.profile = profile

    def versionMatch(self, source, dest=None):
        return _version_matches(
            source, self.source, self.dest, strict=True, dest=dest
        )

    def isProposed(self, tool, source, dest=None):
        """Check if a step can be applied.

        False means already applied or does not apply.
        True means can be applied.
        """
        if not self.versionMatch(source, dest=dest):
            return False
        checker = self.checker
        if checker is None:
            return True
        return checker(tool)


class UpgradeStep(UpgradeEntity):
    """A step to upgrade a component.
    """

    def __init__(self, title, profile, source, dest, desc, handler,
                 checker=None, sortkey=0):
        super().__init__(title, profile, source, dest, desc, checker, sortkey)
        self.handler = handler

    def doStep(self, tool):
        self.handler(tool)


class UpgradeDepends(UpgradeEntity):
    """A specialized upgrade step that re-runs a particular import
    step from the profile.
    """

    def __init__(self, title, profile, source, dest, desc, import_profile=None,
                 import_steps=[], run_deps=False, purge=False, checker=None,
                 sortkey=0):
        super().__init__(title, profile, source, dest, desc, checker, sortkey)
        self.import_profile = import_profile
        self.import_steps = import_steps
        self.run_deps = run_deps
        self.purge = purge

    PROFILE_PREFIX = 'profile-%s'

    def doStep(self, tool):
        if self.import_profile is None:
            profile_id = self.PROFILE_PREFIX % self.profile
        else:
            profile_id = self.PROFILE_PREFIX % self.import_profile
        if self.import_steps:
            for step in self.import_steps:
                tool.runImportStepFromProfile(profile_id, step,
                                              run_dependencies=self.run_deps,
                                              purge_old=self.purge)
        else:
            # if no steps are specified we assume we want to reimport
            # the entire profile
            ign_deps = not self.run_deps
            tool.runAllImportStepsFromProfile(profile_id,
                                              purge_old=self.purge,
                                              ignore_dependencies=ign_deps)


def _registerUpgradeStep(step):
    profile_id = step.profile
    profile_steps = _upgrade_registry.getUpgradeStepsForProfile(profile_id)
    profile_steps[step.id] = step


def _registerNestedUpgradeStep(step, outer_id):
    profile_id = step.profile
    profile_steps = _upgrade_registry.getUpgradeStepsForProfile(profile_id)
    nested_steps = profile_steps.get(outer_id, [])
    nested_steps.append((step.id, step))
    profile_steps[outer_id] = nested_steps


def _extractStepInfo(tool, id, step, source, dest=None):
    """Returns the info data structure for a given step.
    """
    proposed = step.isProposed(tool, source, dest=dest)
    if not proposed and not _version_matches(
            source, step.source, step.dest, dest=dest
    ):
        return
    info = {
        'id': id,
        'step': step,
        'title': step.title,
        'source': step.source,
        'dest': step.dest,
        'description': step.description,
        'proposed': proposed,
        'sortkey': step.sortkey,
    }
    return info


def listProfilesWithUpgrades():
    return _upgrade_registry.keys()


def listUpgradeSteps(tool, profile_id, source, dest=None, quick=False):
    """Lists upgrade steps available from a given version, for a given
    profile id.

    If 'quick' is True, we return a simple boolean True when we have found the
    first matching upgrade step.  This is useful when you only want to know
    if there is at least one upgrade step.  This is used by
    tool.hasUpgradeSteps.
    """
    res = []
    profile_steps = _upgrade_registry.getUpgradeStepsForProfile(profile_id)
    # Optionally limit to a maximum destination.
    if isinstance(dest, str):
        dest = tuple(dest.split('.'))
    for id, step in profile_steps.items():
        if isinstance(step, UpgradeEntity):
            info = _extractStepInfo(tool, id, step, source, dest=dest)
            if info is None:
                continue
            if quick:
                return True
            normsrc = normalize_version(step.source)
            res.append(((normsrc, step.sortkey, info['proposed']), info))
        else:  # nested steps
            nested = []
            outer_proposed = False
            for inner_id, inner_step in step:
                info = _extractStepInfo(
                    tool, inner_id, inner_step, source, dest=dest
                )
                if info is None:
                    continue
                if quick:
                    return True
                nested.append(info)
                outer_proposed = outer_proposed or info['proposed']
            if nested:
                src = nested[0]['source']
                sortkey = nested[0]['sortkey']
                normsrc = normalize_version(src)
                res.append(((normsrc, sortkey, outer_proposed), nested))
    res.sort(key=lambda x: x[0])
    res = [i[1] for i in res]
    return res



##############################################################################
#
# Copyright (c) 2004 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Diff utilities for comparing configurations.
"""

import difflib
import re

from AccessControl.class_init import InitializeClass
from AccessControl.SecurityInfo import ClassSecurityInfo

from .interfaces import SKIPPED_FILES


BLANKS_REGEX = re.compile(b'^\\s*$')


def unidiff(a, b, filename_a=b'original', timestamp_a=b'',
            filename_b=b'modified', timestamp_b=b'', ignore_blanks=False):
    r"""Compare two sequences of lines; generate the resulting delta.

    Each sequence must contain individual single-line strings
    ending with newlines. Such sequences can be obtained from the
    `readlines()` method of file-like objects.  The delta
    generated also consists of newline-terminated strings, ready
    to be printed as-is via the writeline() method of a file-like
    object.

    Note that the last line of a file may *not* have a newline;
    this is reported in the same way that GNU diff reports this.
    *This method only supports UNIX line ending conventions.*

        filename_a and filename_b are used to generate the header,
        allowing other tools to determine what 'files' were used
        to generate this output.

        timestamp_a and timestamp_b, when supplied, are expected
        to be last-modified timestamps to be inserted in the
        header, as floating point values since the epoch.

    """
    if isinstance(a, bytes):
        a = a.splitlines()

    if isinstance(b, bytes):
        b = b.splitlines()

    if isinstance(filename_a, str):
        filename_a = filename_a.encode('utf-8')

    if isinstance(filename_b, str):
        filename_b = filename_b.encode('utf-8')

    if not isinstance(timestamp_a, bytes):
        timestamp_a = str(timestamp_a).encode('utf-8')

    if not isinstance(timestamp_b, bytes):
        timestamp_b = str(timestamp_b).encode('utf-8')

    if ignore_blanks:
        a = [x for x in a if not BLANKS_REGEX.match(x)]
        b = [x for x in b if not BLANKS_REGEX.match(x)]

    return difflib.diff_bytes(difflib.unified_diff, a, b, filename_a,
                              filename_b, timestamp_a, timestamp_b,
                              lineterm=b"")


class ConfigDiff:

    security = ClassSecurityInfo()

    def __init__(self, lhs, rhs, missing_as_empty=False, ignore_blanks=False,
                 skip=SKIPPED_FILES):
        self._lhs = lhs
        self._rhs = rhs
        self._missing_as_empty = missing_as_empty
        self._ignore_blanks = ignore_blanks
        self._skip = skip

    @security.private
    def compareDirectories(self, subdir=None):

        lhs_files = self._lhs.listDirectory(subdir, self._skip)
        if lhs_files is None:
            lhs_files = []

        rhs_files = self._rhs.listDirectory(subdir, self._skip)
        if rhs_files is None:
            rhs_files = []

        added = [f for f in rhs_files if f not in lhs_files]
        removed = [f for f in lhs_files if f not in rhs_files]
        all_files = lhs_files + added
        all_files.sort()

        result = []

        for filename in all_files:

            if subdir is None:
                pathname = filename
            else:
                pathname = f'{subdir}/{filename}'

            if filename not in added:
                isDirectory = self._lhs.isDirectory(pathname)
            else:
                isDirectory = self._rhs.isDirectory(pathname)

            if not self._missing_as_empty and filename in removed:

                if isDirectory:
                    result.append(b'** Directory %s removed\n' %
                                  pathname.encode('utf-8'))
                    result.extend(self.compareDirectories(pathname))
                else:
                    result.append(b'** File %s removed\n' %
                                  pathname.encode('utf-8'))

            elif not self._missing_as_empty and filename in added:

                if isDirectory:
                    result.append(b'** Directory %s added\n' %
                                  pathname.encode('utf-8'))
                    result.extend(self.compareDirectories(pathname))
                else:
                    result.append(b'** File %s added\n' %
                                  pathname.encode('utf-8'))

            elif isDirectory:

                result.extend(self.compareDirectories(pathname))

                if (filename not in added + removed and
                        not self._rhs.isDirectory(pathname)):

                    result.append(b'** Directory %s replaced with a file of '
                                  b'the same name\n' % pathname)

                    if self._missing_as_empty:
                        result.extend(self.compareFiles(filename, subdir))
            else:
                if (filename not in added + removed and
                        self._rhs.isDirectory(pathname)):

                    result.append(b'** File %s replaced with a directory of '
                                  b'the same name\n' % pathname)

                    if self._missing_as_empty:
                        result.extend(self.compareFiles(filename, subdir))

                    result.extend(self.compareDirectories(pathname))
                else:
                    result.extend(self.compareFiles(filename, subdir))

        return result

    @security.private
    def compareFiles(self, filename, subdir=None):

        if subdir is None:
            path = filename
        else:
            path = f'{subdir}/{filename}'

        lhs_file = self._lhs.readDataFile(filename, subdir)
        lhs_time = self._lhs.getLastModified(path)

        if lhs_file is None:
            assert self._missing_as_empty
            lhs_file = b''
            lhs_time = b''

        rhs_file = self._rhs.readDataFile(filename, subdir)
        rhs_time = self._rhs.getLastModified(path)

        if rhs_file is None:
            assert self._missing_as_empty
            rhs_file = b''
            rhs_time = b''

        if lhs_file == rhs_file:
            diff_lines = []
        else:
            diff_lines = unidiff(lhs_file, rhs_file, filename_a=path,
                                 timestamp_a=lhs_time, filename_b=path,
                                 timestamp_b=rhs_time,
                                 ignore_blanks=self._ignore_blanks)
            diff_lines = list(diff_lines)  # generator

        if len(diff_lines) == 0:  # No *real* difference found
            return []

        diff_lines.insert(0, b'Index: %s' % path.encode('utf-8'))
        diff_lines.insert(1, b'=' * 67)

        return diff_lines

    @security.private
    def compare(self):
        return b'\n'.join(self.compareDirectories())


InitializeClass(ConfigDiff)
